# -*- coding: utf-8 -*-
"""
datetime manager module.
"""

from datetime import datetime

import pytz

import pyrin.configuration.services as config_services
import pyrin.utils.datetime as datetime_utils

from pyrin.core.context import Manager


class DateTimeManager(Manager):
    """
    datetime manager class.
    """

    def __init__(self, **options):
        """
        initializes an instance of DateTimeManager.
        """

        super().__init__()

        default_timezone_name = config_services.get('globalization',
                                                    'locale',
                                                    'babel_default_timezone')
        self.__current_timezone = pytz.timezone(default_timezone_name)

    def now(self, timezone=None):
        """
        gets the current datetime based on given timezone name.

        :param str timezone: timezone name to get current datetime based on it.
                             if not provided, defaults to application
                             current timezone.

        :rtype: datetime
        """

        if timezone is None:
            timezone = self.get_current_timezone()
        else:
            timezone = pytz.timezone(timezone)

        return datetime.now(timezone)

    def get_current_timezone(self):
        """
        gets the application current timezone.

        :rtype: tzinfo
        """

        return self.__current_timezone

    def get_current_timezone_name(self):
        """
        gets the application current timezone name.

        :rtype: str
        """

        return self.get_current_timezone().zone

    def normalize(self, value):
        """
        normalizes input datetime value using application current
        timezone and returns it.
        input value should already have a timezone info.

        :param datetime value: value to get normalized.

        :rtype: datetime
        """

        return self.get_current_timezone().normalize(value)

    def localize(self, value):
        """
        localizes input datetime with application current timezone.
        input value should not have a timezone info.

        :param datetime value: value to be localized.

        :rtype: datetime
        """

        return self.get_current_timezone().localize(value)

    def _add_timezone_info(self, value):
        """
        adds the application current timezone info into input
        value if it has no timezone info.

        :param datetime value: value to add timezone info into it.

        :rtype: datetime
        """

        localized_value = value
        if value.tzinfo is None:
            localized_value = self.get_current_timezone().localize(value)

        return localized_value

    def to_datetime_string(self, value):
        """
        gets the datetime string representation of input value.
        if the value has no timezone info, it adds the application
        current timezone info into it.
        example: `2015-12-24T22:40:15+01:00`

        :param datetime value: input object to be converted.

        :rtype: str
        """

        localized_value = self._add_timezone_info(value)
        return datetime_utils.to_datetime_string(localized_value)

    def to_date_string(self, value):
        """
        gets the date string representation of input value.
        example: `2015-12-24`

        :param Union[datetime, date] value: input object to be converted.

        :rtype: str
        """

        return datetime_utils.to_date_string(value)

    def to_time_string(self, value):
        """
        gets the time string representation of input value.
        if the value has no timezone info, it adds the application
        current timezone info into it.
        example: `23:40:15`

        :param Union[datetime, time] value: input object to be converted.

        :rtype: str
        """

        localized_value = value
        if isinstance(value, datetime):
            localized_value = self._add_timezone_info(value)
        return datetime_utils.to_time_string(localized_value)

    def to_datetime(self, value):
        """
        converts the input value to it's equivalent python datetime.

        :param str value: string representation of datetime to be converted.

        :rtype: datetime
        """

        converted_datetime = datetime_utils.to_datetime(value)
        return self._add_timezone_info(converted_datetime)

    def to_date(self, value):
        """
        converts the input value to it's equivalent python date.

        :param str value: string representation of date to be converted.

        :rtype: date
        """

        return datetime_utils.to_date(value)

    def to_time(self, value):
        """
        converts the input value to it's equivalent python time.

        :param str value: string representation of time to be converted.

        :rtype: time
        """

        return datetime_utils.to_time(value)

    def timezone_exists(self, timezone_name):
        """
        gets a value indicating that a timezone with the given name exists.

        :param str timezone_name: timezone name to check for existence.

        :rtype: bool
        """

        return timezone_name in pytz.all_timezones_set

    def get_current_timestamp(self, date_sep='-', main_sep=' ',
                              time_sep=':', timezone=None):
        """
        gets the current timestamp with specified separators based on given timezone.

        :param Union[str, None] date_sep: a separator to put between date elements.
        :param Union[str, None] main_sep: a separator to put between date and time part.
        :param Union[str, None] time_sep: a separator to put between time elements.

        :param Union[str, None] timezone: timezone name to get current timestamp
                                          based on it. if not provided, defaults
                                          to application current timezone.

        :rtype: str
        """

        if date_sep is None:
            date_sep = ''
        if main_sep is None:
            main_sep = ''
        if time_sep is None:
            time_sep = ''

        current = self.now(timezone)
        return '{year}{date_sep}{month}{date_sep}{day}{main_sep}' \
               '{hour}{time_sep}{minute}{time_sep}{second}'.format(
                year=str(current.year).zfill(4),
                month=str(current.month).zfill(2),
                day=str(current.day).zfill(2),
                hour=str(current.hour).zfill(2),
                minute=str(current.minute).zfill(2),
                second=str(current.second).zfill(2),
                date_sep=date_sep,
                main_sep=main_sep,
                time_sep=time_sep)



# -*- coding: utf-8 -*-
"""
serializer base module.
"""

from threading import Lock

from pyrin.core.context import CoreObject
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.utils.singleton import MultiSingletonMeta


class SerializerSingletonMeta(MultiSingletonMeta):
    """
    serializer singleton meta class.
    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class SerializerBase(CoreObject, metaclass=SerializerSingletonMeta):
    """
    serializer base class.
    all application serializers must inherit from this.
    """

    @classmethod
    def serialize(cls, value, **options):
        """
        serializes the given value.

        :param object value: value to be serialized.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: serialized object.
        """

        raise CoreNotImplementedError()

    @classmethod
    def serialize_list(cls, values, **options):
        """
        serializes the given list of values.

        :param list[object] values: values to be serialized.

        :returns: list of serialized objects.
        :rtype: list
        """

        result = []
        if values is None or len(values) == 0:
            return result

        for value in values:
            result.append(cls.serialize(value, **options))

        return result



# -*- coding: utf-8 -*-
"""
utils exceptions module.
"""

from pyrin.core.exceptions import CoreException


class UtilsException(CoreException):
    """
    utils exception.
    """
    pass


class InputNotCallableError(UtilsException):
    """
    input not callable error.
    """
    pass


class ConfigurationFileNotFoundError(UtilsException):
    """
    configuration file not found error.
    """
    pass


class InvalidSchemaNameError(UtilsException):
    """
    invalid schema name error.
    """
    pass


class InvalidTableNameError(UtilsException):
    """
    invalid table name error.
    """
    pass


class InvalidColumnNameError(UtilsException):
    """
    invalid column name error.
    """
    pass


class InvalidRowResultFieldsAndValuesError(UtilsException):
    """
    invalid row result fields and values error.
    """
    pass


class FieldsAndValuesCountMismatchError(UtilsException):
    """
    fields and values count mismatch error.
    """
    pass


class DictKeyPrefixIsNotProvidedError(UtilsException):
    """
    dict key prefix is not provided error.
    """
    pass



# -*- coding: utf-8 -*-
"""
configuration store module.
"""

import os

import pyrin.converters.deserializer.services as deserializer_services
import pyrin.utils.configuration as config_utils

from pyrin.utils.exceptions import ConfigurationFileNotFoundError as UtilsFileNotFoundError
from pyrin.core.context import CoreObject, DTO
from pyrin.utils.custom_print import print_warning
from pyrin.utils.dictionary import change_key_case
from pyrin.configuration.exceptions import ConfigurationFileNotFoundError, \
    ConfigurationStoreKeyNotFoundError, ConfigurationStoreSectionNotFoundError, \
    ConfigurationStoreDuplicateKeyError, ConfigurationEnvironmentVariableNotFoundError, \
    InvalidConfigurationEnvironmentVariableValueError, \
    ConfigurationStoreAttributeNotFoundException


class ConfigStore(CoreObject):
    """
    config store class.
    all configurations will be stored with lowercase keys.
    """

    ACTIVE_SECTION_NAME = 'active'
    SELECTED_SECTION_NAME = 'selected'

    def __init__(self, name, config_file_path, **options):
        """
        initializes a new ConfigStore.

        :param str name: config store name.
        :param str config_file_path: full path of config file.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        """

        super().__init__()

        self._configs = DTO()
        self._name = name
        self._config_file_path = config_file_path
        self._load(**options)

    def _load(self, **options):
        """
        loads configurations from config file path.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        """

        try:
            self._configs = config_utils.load(self._config_file_path,
                                              deserializer_services.deserialize)
        except UtilsFileNotFoundError as error:
            raise ConfigurationFileNotFoundError(error) from error

        self._sync_with_env(**options)

    def reload(self, **options):
        """
        reloads configuration from it's physical file path.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        """

        self._configs.clear()
        self._load(**options)

    def get(self, section, key, **options):
        """
        gets the value of given key from provided section.

        :param str section: config section name.
        :param str key: config key to get it's value.

        :keyword object default_value: default value if key not present in config section.
                                       if not provided, error will be raised.

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.
        """

        section_data = self.get_section(section, **options)
        if key in section_data.keys():
            return section_data[key]

        if 'default_value' not in options.keys():
            raise ConfigurationStoreKeyNotFoundError('Key [{key}] not found in section '
                                                     '[{section}] from config store [{name}].'
                                                     .format(key=key,
                                                             section=section,
                                                             name=self._name))

        return options.get('default_value')

    def get_active(self, key, **options):
        """
        gets the value of given key from active section of this config store.
        if this store does not have an active section, it raises an error.

        :param str key: config key to get it's value.

        :keyword object default_value: default value if key not present in config section.
                                       if not provided, error will be raised.

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.
        """

        active_section = self.get_active_section_name()
        return self.get(active_section, key, **options)

    def get_section_names(self, **options):
        """
        gets all available section names of config store.

        :rtype: list[str]
        """

        return self._configs.keys()

    def get_section(self, section, **options):
        """
        gets all key/values stored in given section.

        :param str section: section name.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                        not found error.

        :rtype: dict
        """

        if section not in self._configs.keys():
            raise ConfigurationStoreSectionNotFoundError('Section [{section}] not found in '
                                                         'config store [{name}].'
                                                         .format(section=section,
                                                                 name=self._name))

        result = self._configs.get(section)

        return self._change_key_case(result, **options)

    def get_section_keys(self, section, **options):
        """
        gets all available keys in given section.

        :param str section: section name.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                        not found error.

        :rtype: list[str]
        """

        return self.get_section(section, **options).keys()

    def get_all(self, **options):
        """
        gets all available key/values from different sections of
        this config store in a flat dict, eliminating the sections.
        note that if there are same key names in different
        sections, it raises an error to prevent overwriting values.
        also note that if this config store contains `active` section,
        then the result of `get_active_section` method would be returned.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreDuplicateKeyError: configuration store duplicate key error.

        :rtype: dict
        """

        # trying to get active config if available.
        active_config = None
        try:
            active_config = self.get_active_section(**options)
        except ConfigurationStoreAttributeNotFoundException:
            pass

        if active_config is not None:
            return active_config

        flat_dict = DTO()
        for section_data in self._get_sections(**options):
            if isinstance(section_data, dict):
                for key, value in section_data.items():
                    if key in flat_dict.keys():
                        raise ConfigurationStoreDuplicateKeyError('Key [{key}] is available '
                                                                  'in multiple sections of '
                                                                  'config store [{name}].'
                                                                  .format(key=key,
                                                                          name=self._name))
                    flat_dict[key] = value

        return flat_dict

    def get_file_path(self, **options):
        """
        gets config file path of this config store.

        :rtype: str
        """

        return self._config_file_path

    def _change_key_case(self, value, **options):
        """
        returns a copy of input dict with all it's keys
        and nested keys cases modified using given converter.
        if converter is not provided, it returns the inputted dict.

        :param dict value: dict to change it's keys cases.

        :keyword callable converter: a callable to use as case converter.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :rtype: dict
        """

        converter = options.get('converter', None)
        if converter is not None:
            return change_key_case(value, converter)

        return value

    def get_active_section(self, **options):
        """
        gets the active section available in related config file.
        this method gets the section that it's name is under [active]
        section, for example:

        [active]
        selected: production

        [production]
        id: 123
        name: prod

        [development]
        id: 233
        name: dev

        this will return all key/values available under [production].
        if the config store has not an [active] section, this method
        raises an error.

        :keyword callable converter: a callable to use as case converter.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.

        :rtype: dict
        """

        active_section = self.get_active_section_name()
        return self.get_section(active_section, **options)

    def _get_sections(self, **options):
        """
        gets a list of dictionaries containing all key/values
        for every section of this config store.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :rtype: list[dict]
        """

        return [self.get_section(section, **options)
                for section in self.get_section_names(**options)]

    def _get_from_env(self, key, **options):
        """
        gets the value of given key from environment variable if available.
        otherwise may raise an exception or ignore it depending on `silent` keyword.

        :param str key: key to get it's value from environment variable.

        :keyword bool silent: indicates that if an environment variable for the
                              given key not found, ignore it and return None, otherwise
                              raise an error. defaults to True.

        :raises ConfigurationEnvironmentVariableNotFoundError: configuration environment
                                                               variable not found error.

        :raises InvalidConfigurationEnvironmentVariableValueError: invalid configuration
                                                                   environment variable
                                                                   value error.

        :rtype: str
        """

        value = os.environ.get(key)
        silent = options.get('silent', True)
        if value is None:
            message = 'Configuration environment variable [{key}] not found.'.format(key=key)
            if silent is not True:
                raise ConfigurationEnvironmentVariableNotFoundError(message)
            else:
                print_warning(message)

        if value is not None and len(value.strip()) == 0:
            message = 'Configuration environment variable ' \
                      '[{key}] has an invalid value.'.format(key=key)
            if silent is not True:
                raise InvalidConfigurationEnvironmentVariableValueError(message)
            else:
                print_warning(message)

        return value

    def _sync_with_env(self, **options):
        """
        synchronizes all keys with None value of this config store
        with environment variables with the same name if available.

        :keyword bool silent: indicates that if an environment variable for the
                              config key not found, ignore it and return None, otherwise
                              raise an error. defaults to True.
        """

        for section_name in self.get_section_names():
            section = self.get_section(section_name)
            for key, value in section.items():
                if value is None:
                    env_value = self._get_from_env(key, **options)
                    converted_value = deserializer_services.deserialize(env_value)
                    self._configs[section_name][key] = converted_value

    def get_active_section_name(self):
        """
        gets the active section name of this config store if available.
        if the store does not have an active section, it raises an error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.

        :rtype: str
        """

        return self.get(self.ACTIVE_SECTION_NAME, self.SELECTED_SECTION_NAME)

    def __str__(self):
        return self._name



# -*- coding: utf-8 -*-
"""
configuration manager module.
"""

import os

import pyrin.application.services as application_services

from pyrin.configuration.store import ConfigStore
from pyrin.core.context import Context, Manager
from pyrin.utils.custom_print import print_warning
from pyrin.configuration.exceptions import ConfigurationStoreExistedError, \
    ConfigurationStoreNotFoundError, ConfigurationFileNotFoundError


class ConfigurationManager(Manager):
    """
    configuration manager class.
    """

    def __init__(self, **options):
        """
        initializes an instance of ConfigurationManager.
        """

        super().__init__()

        self._config_stores = Context()
        self._settings_path = application_services.get_settings_path()
        self._default_settings_path = application_services.get_default_settings_path()

    def _add_config_store(self, name, file_path, **options):
        """
        adds a new config store for given file with the specified name.

        :param str name: config store name.
        :param str file_path: config file full path.

        :keyword bool ignore_on_existed: specifies that it should not raise an
                                         error if a config store with given name
                                         has been already loaded.
                                         defaults to False if not provided.

        :raises ConfigurationStoreExistedError: configuration store existed error.
        """

        ignore = options.get('ignore_on_existed', False)
        if name in self._config_stores:
            if ignore is not True:
                raise ConfigurationStoreExistedError('Config store with name [{name}] already '
                                                     'existed, config file names must be unique.'
                                                     .format(name=name))
            else:
                print_warning('Config store with name [{name}] already '
                              'existed, it will be ignored.'
                              .format(name=name))
                return

        self._config_stores[name] = ConfigStore(name, file_path, **options)

    def _is_config_file(self, file_name):
        """
        gets a value indicating that given file name belongs to a config file.

        :param str file_name: file name.

        :rtype: bool
        """

        return file_name.endswith('.config')

    def load_configuration(self, name, **options):
        """
        loads the given configuration if relevant file is
        available in settings path.

        :param str name: configuration name.

        :keyword bool silent: specifies that if a related configuration file
                              for the given name not found, ignore it.
                              otherwise raise an error. defaults to False.

        :keyword bool ignore_on_existed: specifies that it should not raise an
                                         error if a config store with given name
                                         has been already loaded.
                                         defaults to False if not provided.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        :raises ConfigurationStoreExistedError: configuration store existed error.
        """

        file_path = self._get_relevant_file_path(name, **options)
        if file_path is not None:
            self._add_config_store(name, file_path, **options)

    def load_configurations(self, *names, **options):
        """
        loads the given configurations if relevant files is
        available in settings path.

        :param str names: configuration names as arguments.

        :keyword bool silent: specifies that if a related configuration file
                              for any of the given names not found, ignore it.
                              otherwise raise an error. defaults to False.

        :keyword bool ignore_on_existed: specifies that it should not raise an
                                         error if a config store with given name
                                         has been already loaded.
                                         defaults to False if not provided.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        :raises ConfigurationStoreExistedError: configuration store existed error.
        """

        for single_name in names:
            self.load_configuration(single_name, **options)

    def _get_relevant_file_path(self, name, **options):
        """
        gets the relevant file path to specified config name in settings folder.

        :param str name: config store name.

        :keyword bool silent: specifies that if a related configuration file
                              for the given store name not found, ignore it.
                              otherwise raise an error. defaults to False.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        """

        file_path = self._try_get_relevant_file_path(self._settings_path, name)
        if file_path is None:
            file_path = self._try_get_relevant_file_path(self._default_settings_path, name)

        if file_path is None:
            silent = options.get('silent', False)
            if silent is not True:
                raise ConfigurationFileNotFoundError('Config name [{name}] does not '
                                                     'have any related configuration '
                                                     'file in application settings.'
                                                     .format(name=name))
        return file_path

    def _try_get_relevant_file_path(self, settings_path, name):
        """
        tries to get the relevant file path to specified
        config name in given settings path.
        it may return a path if finds the file or return None if not found.

        :param str settings_path: settings path to search in for file name.
        :param str name: config store name.

        :rtype: str
        """

        if not os.path.isdir(settings_path):
            return None

        files = os.listdir(settings_path)
        for single_file in files:
            single_file_name = os.path.splitext(single_file)[0]
            if single_file_name == name and \
               self._is_config_file(single_file):
                return os.path.abspath(os.path.join(settings_path, single_file))

        return None

    def reload(self, store_name, **options):
        """
        reloads the configuration store from it's relevant file.

        :param str store_name: config store name to be reloaded.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.
        """

        self._get_config_store(store_name).reload(**options)

    def get_file_path(self, store_name, **options):
        """
        gets the configuration file path for given config store.

        :param str store_name: config store name to get it's file path.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.
        """

        if store_name in self._config_stores.keys():
            return self._get_config_store(store_name).get_file_path(**options)

        return self._get_relevant_file_path(store_name, **options)

    def get(self, store_name, section, key, **options):
        """
        gets the value of specified key from provided section of given config store.

        :param str store_name: config store name.
        :param str section: config section name.
        :param str key: config key to get it's value.

        :keyword object default_value: default value if key not present in config section.
                                       if not provided, error will be raised.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.
        """

        return self._get_config_store(store_name).get(section, key, **options)

    def get_active(self, store_name, key, **options):
        """
        gets the value of given key from active section of given config store.
        if this store does not have an active section, it raises an error.

        :param str store_name: config store name.
        :param str key: config key to get it's value.

        :keyword object default_value: default value if key not present in config section.
                                       if not provided, error will be raised.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.
        """

        return self._get_config_store(store_name).get_active(key, **options)

    def get_section_names(self, store_name, **options):
        """
        gets all available section names of given config store.

        :param str store_name: config store name.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :rtype: list[str]
        """

        return self._get_config_store(store_name).get_section_names(**options)

    def get_section(self, store_name, section, **options):
        """
        gets all key/values stored in given section of specified config store.

        :param str store_name: config store name.
        :param str section: section name.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                        not found error.

        :rtype: dict
        """

        return self._get_config_store(store_name).get_section(section, **options)

    def get_section_keys(self, store_name, section, **options):
        """
        gets all available keys in given section of specified config store.

        :param str store_name: config store name.
        :param str section: section name.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                        not found error.

        :rtype: list[str]
        """

        return self._get_config_store(store_name).get_section_keys(section, **options)

    def get_all(self, store_name, **options):
        """
        gets all available key/values from different sections of
        given config store in a flat dict, eliminating the sections.
        note that if there are same key names in different
        sections, it raises an error to prevent overwriting values.
        also note that if the config store contains `active` section,
        then the result of `get_active_section` method would be returned.

        :param str store_name: config store name.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreNotFoundError: configuration store not found error.
        :raises ConfigurationStoreDuplicateKeyError: configuration store duplicate key error.

        :rtype: dict
        """

        return self._get_config_store(store_name).get_all(**options)

    def get_active_section(self, store_name, **options):
        """
        gets the active section available in given config store.
        this method gets the section that it's name is under [active]
        section, for example:

        [active]
        selected: production

        [production]
        id: 123
        name: prod

        [development]
        id: 233
        name: dev

        this will return all key/values available under [production].
        if the config store has not an [active] section, this method
        raises an error.

        :param str store_name: config store name.

        :keyword callable converter: a callable to use as case converter.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                        not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.

        :rtype: dict
        """

        return self._get_config_store(store_name).get_active_section(**options)

    def get_active_section_name(self, store_name):
        """
        gets the active section name of given config store if available.
        if the store does not have an active section, it raises an error.

        :param str store_name: config store name.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.

        :rtype: str
        """

        return self._get_config_store(store_name).get_active_section_name()

    def _get_config_store(self, name):
        """
        gets the specified config store.

        :param str name: config store name.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :rtype: ConfigStore
        """

        if name not in self._config_stores.keys():
            raise ConfigurationStoreNotFoundError('Config store [{name}] not found.'
                                                  .format(name=name))

        return self._config_stores[name]



# -*- coding: utf-8 -*-
"""
database decorators module.
"""

from functools import update_wrapper

import pyrin.database.services as database_services

from pyrin.database.services import get_current_store


def session_factory(*args, **kwargs):
    """
    decorator to register a database session factory.

    :param object args: session factory class constructor arguments.
    :param object kwargs: session factory class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           session factory with the same name, replace it with
                           the new one, otherwise raise an error. defaults to False.

    :raises InvalidSessionFactoryTypeError: invalid session factory type error.
    :raises DuplicatedSessionFactoryError: duplicated session factory error.

    :returns: session factory class.

    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available database session factories.

        :param type cls: session factory class.

        :returns: session factory class.

        :rtype: type
        """

        instance = cls(*args, **kwargs)
        database_services.register_session_factory(instance, **kwargs)

        return cls

    return decorator


def bind(name, **options):
    """
    decorator to bind a model class to a database.

    :param str name: bind name to associate with the model.

    :raises InvalidEntityTypeError: invalid entity type error.

    :returns: model class.

    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given model class and binds it with the specified database.

        :param type cls: model class.

        :returns: model class.

        :rtype: type
        """

        database_services.register_bind(cls, name, **options)

        return cls

    return decorator


def database_hook():
    """
    decorator to register a database hook.

    :raises InvalidDatabaseHookTypeError: invalid database hook type error.

    :returns: database hook class.

    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available database hooks.

        :param type cls: database hook class.

        :returns: database hook class.

        :rtype: type
        """

        instance = cls()
        database_services.register_hook(instance)

        return cls

    return decorator


def atomic(func):
    """
    decorator to make a function execution atomic.
    meaning that before starting the execution of the function, a sub-transaction
    will be started, and after the completion of that function, if it was successful,
    the sub-transaction will be committed or if it was not successful the sub-transaction
    will be rolled-back without the consideration or affecting the parent transaction
    which by default is scoped to request.

    :param callable func: function.

    :returns: function result.
    """

    def decorator(*args, **kwargs):
        """
        decorates the given function and makes its execution atomic.

        :param object args: function arguments.
        :param object kwargs: function keyword arguments.

        :returns: function result.
        """

        store = get_current_store()
        transaction = store.begin_nested()
        try:
            result = func(*args, **kwargs)
            transaction.commit()
            return result
        except Exception as ex:
            if transaction is not None:
                transaction.rollback()
            raise ex

    return update_wrapper(decorator, func)



# -*- coding: utf-8 -*-
"""
core globals module.
"""

from flask_babel import gettext

from pyrin.core.context import CoreObject


# this value should be used as `None`, where the `None` itself has a meaning.
NULL = CoreObject()

# this value should be used where we need to reference class type of None objects.
NONE_TYPE = type(None)

# this value should be used where we want to determine if an object is a list type or not.
LIST_TYPES = (list, tuple, set)

# this method should be used for localizable strings.
_ = gettext



# -*- coding: utf-8 -*-
"""
security utils key_helper module.
"""

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization

import pyrin.configuration.services as config_services

from pyrin.settings.static import APPLICATION_ENCODING


def generate_rsa_key(**options):
    """
    gets a pair of public/private rsa keys.

    :keyword int length: key length in bits.

    :returns: tuple(str public_key, str private_key)

    :rtype: tuple(str, str)
    """

    key_size = options.get('length', None)
    if key_size is None:
        key_size = config_services.get('security', 'general', 'rsa_default_key_length')

    key = rsa.generate_private_key(public_exponent=65537,
                                   key_size=key_size,
                                   backend=default_backend())

    private = key.private_bytes(encoding=serialization.Encoding.PEM,
                                format=serialization.PrivateFormat.TraditionalOpenSSL,
                                encryption_algorithm=serialization.NoEncryption())

    public_part = key.public_key()
    public = public_part.public_bytes(encoding=serialization.Encoding.PEM,
                                      format=serialization.PublicFormat.SubjectPublicKeyInfo)

    return public.decode(APPLICATION_ENCODING), private.decode(APPLICATION_ENCODING)


def load_rsa_key(public_pem, private_pem, **options):
    """
    loads public/private rsa key objects from provided value.

    :param str public_pem: public key content to load from.
    :param str private_pem: private key content to load from.

    :returns: tuple(public_key, private_key)

    :rtype: tuple(object, object)
    """

    private_key = serialization.load_pem_private_key(private_pem.encode(APPLICATION_ENCODING),
                                                     password=None,
                                                     backend=default_backend())

    public_key = serialization.load_pem_public_key(public_pem.encode(APPLICATION_ENCODING),
                                                   backend=default_backend())

    return public_key, private_key



# -*- coding: utf-8 -*-
"""
database orm sql package.
"""

from pyrin.packaging.context import Package


class DatabaseORMSQLPackage(Package):
    """
    database orm sql package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
locale babel package.
"""

from pyrin.packaging.context import Package


class LocaleBabelPackage(Package):
    """
    locale babel package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'globalization.locale.babel.component'



# -*- coding: utf-8 -*-
"""
utils paging module.
"""

import pyrin.configuration.services as config_services


def inject_limit(options):
    """
    injects default limit parameter into given dict.

    :param dict options: options dict to inject limit into it.
    """

    inject_custom_limit(config_services.get('database', 'paging', 'limit'), options)


def inject_custom_limit(limit, options):
    """
    injects given limit parameter into given dict.

    :param int limit: limit value to be set in given dict.
    :param dict options: options dict to inject limit into it.
    """

    options.update(__limit__=limit)


def extract_limit(options):
    """
    extracts and gets limit parameter from given dict.
    if `__limit__` key is not available in provided
    dict, it gets the value from `database.config` file `paging` section.

    :param dict options: options dict to get limit from it.

    :rtype: int
    """

    return options.get('__limit__', config_services.get('database', 'paging', 'limit'))



# -*- coding: utf-8 -*-
"""
deserializer datetime module.
"""

import pyrin.globalization.datetime.services as datetime_services

from pyrin.converters.deserializer.handlers.base import StringPatternDeserializerBase
from pyrin.converters.deserializer.decorators import deserializer
from pyrin.core.globals import NULL
from pyrin.utils.datetime import DEFAULT_DATE_TIME_ISO_REGEX, DEFAULT_DATE_ISO_REGEX, \
    DEFAULT_TIME_ISO_REGEX, DEFAULT_TIME_NO_TIMEZONE_REGEX


@deserializer()
class DateDeserializer(StringPatternDeserializerBase):
    """
    date deserializer class.
    """

    def __init__(self, **options):
        """
        creates an instance of DateDeserializer.

        :keyword list[tuple(Pattern, int)] accepted_formats: a list of custom accepted formats
                                                             and their length for date
                                                             deserialization.

        :type accepted_formats: list[tuple(Pattern format, int length)]
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.
        returns `NULL` object if deserialization fails.

        :param str value: value to be deserialized.

        :rtype: date
        """

        deserializable, pattern = self.is_deserializable(value, **options)
        if not deserializable:
            return NULL

        value = value.strip()
        converted_date = None

        try:
            converted_date = datetime_services.to_date(value)
            if converted_date is not None:
                return converted_date

            return NULL
        except Exception:
            return NULL

    def get_default_formats(self):
        """
        gets default accepted formats that this
        deserializer could deserialize value from.

        :return: list(tuple(Pattern format, int length))

        :rtype: list(tuple(Pattern, int))
        """

        return [(DEFAULT_DATE_ISO_REGEX, 10)]


@deserializer()
class TimeDeserializer(StringPatternDeserializerBase):
    """
    time deserializer class.
    """

    def __init__(self, **options):
        """
        creates an instance of TimeDeserializer.

        :keyword list[tuple(Pattern, int)] accepted_formats: a list of custom accepted formats
                                                             and their length for time
                                                             deserialization.

        :type accepted_formats: list[tuple(Pattern format, int length)]
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.
        returns `NULL` object if deserialization fails.

        :param str value: value to be deserialized.

        :rtype: time
        """

        deserializable, pattern = self.is_deserializable(value, **options)
        if not deserializable:
            return NULL

        value = value.strip()
        converted_time = None

        try:
            converted_time = datetime_services.to_time(value)
            if converted_time is not None:
                return converted_time

            return NULL
        except Exception:
            return NULL

    def get_default_formats(self):
        """
        gets default accepted formats that this
        deserializer could deserialize value from.

        :return: list(tuple(Pattern format, int length))

        :rtype: list(tuple(Pattern, int))
        """

        return [(DEFAULT_TIME_ISO_REGEX, 14),
                (DEFAULT_TIME_NO_TIMEZONE_REGEX, 8)]


@deserializer()
class DateTimeDeserializer(StringPatternDeserializerBase):
    """
    datetime deserializer class.
    """

    def __init__(self, **options):
        """
        creates an instance of DateTimeDeserializer.

        :keyword list[tuple(Pattern, int)] accepted_formats: a list of custom accepted formats
                                                             and their length for datetime
                                                             deserialization.

        :type accepted_formats: list[tuple(Pattern format, int length)]
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.
        returns `NULL` object if deserialization fails.

        :param str value: value to be deserialized.

        :rtype: datetime
        """

        deserializable, pattern = self.is_deserializable(value, **options)
        if not deserializable:
            return NULL

        value = value.strip()
        converted_datetime = None

        try:
            converted_datetime = datetime_services.to_datetime(value)
            if converted_datetime is not None:
                return converted_datetime

            return NULL
        except Exception:
            return NULL

    def get_default_formats(self):
        """
        gets default accepted formats that this
        deserializer could deserialize value from.

        :return: list(tuple(Pattern format, int length))

        :rtype: list(tuple(Pattern, int))
        """

        return [(DEFAULT_DATE_TIME_ISO_REGEX, 25)]



# -*- coding: utf-8 -*-
"""
model base module.
"""

import inspect

from sqlalchemy.ext.declarative import as_declarative
from sqlalchemy.orm import class_mapper, ColumnProperty

import pyrin.database.services as database_services

from pyrin.core.context import CoreObject, DTO
from pyrin.database.model.exceptions import ColumnNotExistedError, EntityNotHashableError


@as_declarative(constructor=None)
class CoreEntity(CoreObject):
    """
    core entity class.
    it will be used as a base class for all models.
    """

    # holds the table name in database.
    __tablename__ = None

    # holds the extra arguments for table.
    # for example:
    # __table_args__ = {'schema': 'database_name.schema_name',
    #                   'extend_existing': True}
    __table_args__ = None

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of CoreDeclarative.
        note that this method will only be called on user code, meaning
        that results returned by orm from database will not call `__init__`
        of each entity.

        it also sets attributes on the constructed instance using the
        names and values in `kwargs`. note that only keys that are present
        as mapped columns of the instance's class are allowed.

        :raises ColumnNotExistedError: column not existed error.
        """

        super().__init__()

        self._set_name(self.__class__.__name__)
        self.from_dict(False, **kwargs)

    def __eq__(self, other):
        if isinstance(other, self._get_root_base_class()):
            if self.primary_key() is not None:
                return self.primary_key() == other.primary_key()
            else:
                return self is other

        return False

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        if self.primary_key() is None:
            raise EntityNotHashableError('Entity [{entity}] does not have '
                                         'a primary key so it is not hashable.'
                                         .format(entity=self))

        return hash('{root_base}.{pk}'.format(root_base=self._get_root_base_class(),
                                              pk=self.primary_key()))

    def __repr__(self):
        return '<{module}.{name} [{pk}]>'.format(module=self.__module__,
                                                 name=self.get_name(),
                                                 pk=self.primary_key())

    def _get_root_base_class(self):
        """
        gets root base class of this entity and caches it.
        root base class is the class which is direct subclass
        of CoreEntity in inheritance hierarchy.
        for example: {Base -> CoreEntity, A -> Base, B -> A}
        root base class of A, B and Base is Base class.

        :rtype: CoreEntity
        """

        if getattr(self, '_root_base_class', None) is None:
            bases = inspect.getmro(type(self))
            base_entity_index = bases.index(CoreEntity) - 1
            self.__class__._root_base_class = bases[base_entity_index]

        return self.__class__._root_base_class

    def _set_all_columns(self, columns):
        """
        sets all column names attribute for this class.

        :param tuple[str] columns: column names.
        """

        if getattr(self, '_all_columns', None) is None:
            self.__class__._all_columns = DTO()

        self.__class__._all_columns[type(self)] = columns

    def _set_exposed_columns(self, columns):
        """
        sets exposed column names attribute for this class.

        :param tuple[str] columns: column names.
        """

        if getattr(self, '_exposed_columns', None) is None:
            self.__class__._exposed_columns = DTO()

        self.__class__._exposed_columns[type(self)] = columns

    def save(self):
        """
        saves the current entity.
        """

        database_services.get_current_store().add(self)
        return self

    def update(self, **kwargs):
        """
        updates the current entity with given values.
        """

        self.from_dict(**kwargs)
        return self.save()

    def delete(self):
        """
        deletes the current entity.
        """

        database_services.get_current_store().delete(self)

    def primary_key(self):
        """
        gets the primary key value of this entity.

        note that the returning value of this method will be used as a way
        to compare two different entities of the same type. so if your table
        does not have a primary key, you could either not implement this method
        and leave comparison to base in the form of `self is other` or you could
        implement another logic in this method to make comparisons possible and
        correct. the returning value of this method must be hashable.

        :rtype: object
        """

        return None

    def all_columns(self):
        """
        gets all column names of entity.
        column names will be calculated once and cached.

        :returns: tuple[str]
        :rtype: tuple
        """

        columns = self._get_all_columns()
        if columns is None:
            all_columns = tuple(prop.key for prop in class_mapper(type(self)).iterate_properties
                                if isinstance(prop, ColumnProperty))
            self._set_all_columns(all_columns)
            columns = self._get_all_columns()

        return columns

    def exposed_columns(self):
        """
        gets exposed column names of entity, which
        are those that have `exposed=True`.
        column names will be calculated once and cached.

        :returns: tuple[str]
        :rtype: tuple
        """

        columns = self._get_exposed_columns()
        if columns is None:
            exposed_columns = tuple(prop.key for prop in class_mapper(type(self)).
                                    iterate_properties if isinstance(prop, ColumnProperty)
                                    and prop.columns[0].exposed is True)
            self._set_exposed_columns(exposed_columns)
            columns = self._get_exposed_columns()

        return columns

    def _get_all_columns(self):
        """
        gets all column names of entity.
        returns None if not found.

        :returns: tuple[str]
        :rtype: tuple
        """

        columns = getattr(self, '_all_columns', None)
        if columns is not None:
            return columns.get(type(self), None)

        return None

    def _get_exposed_columns(self):
        """
        gets exposed column names of entity, which
        are those that have `exposed=True`.
        returns None if not found.

        :returns: tuple[str]
        :rtype: tuple
        """

        columns = getattr(self, '_exposed_columns', None)
        if columns is not None:
            return columns.get(type(self), None)

        return None

    def to_dict(self, exposed_only=True):
        """
        converts the entity into a dict and returns it.
        the result dict only contains the exposed columns of
        the entity which are those that their `exposed` attribute
        is set to True.

        :param bool exposed_only: if set to False, it returns all
                                  columns of the entity as dict.
                                  if not provided, defaults to True.

        :rtype: dict
        """

        columns_collection = self.exposed_columns
        if exposed_only is False:
            columns_collection = self.all_columns

        result = DTO()
        for col in columns_collection():
            result[col] = getattr(self, col)

        return result

    def from_dict(self, silent_on_invalid_column=True, **kwargs):
        """
        updates the column values of the entity from those
        values that are available in input keyword arguments.

        :keyword bool silent_on_invalid_column: specifies that if a key is not available
                                                in entity columns, do not raise an error.
                                                defaults to True if not provided.

        :raises ColumnNotExistedError: column not existed error.
        """

        all_columns = self.all_columns()
        for key, value in kwargs.items():
            if key in all_columns:
                setattr(self, key, value)
            else:
                if silent_on_invalid_column is False:
                    raise ColumnNotExistedError('Entity [{entity}] does not have '
                                                'a column named [{column}].'
                                                .format(entity=self,
                                                        column=key))

    @classmethod
    def table_name(cls):
        """
        gets the table name that this entity represents in database.

        :rtype: str
        """

        return cls.__tablename__

    @classmethod
    def table_schema(cls):
        """
        gets the table schema that this entity represents in database.
        it might be `None` if schema has not been set for this entity.

        :rtype: str
        """

        schema = None
        if cls.__table_args__ is not None:
            schema = cls.__table_args__.get('schema', None)

        return schema

    @classmethod
    def table_fullname(cls):
        """
        gets the table fullname that this entity represents in database.
        fullname is `schema.table_name` if schema is available, otherwise it
        defaults to `table_name`.

        :rtype: str
        """

        schema = cls.table_schema()
        name = cls.table_name()

        if schema is not None:
            return '{schema}.{name}'.format(schema=schema, name=name)
        else:
            return name



# -*- coding: utf-8 -*-
"""
database request_scoped session factory module.
"""

from sqlalchemy.orm import scoped_session, sessionmaker

import pyrin.configuration.services as config_services
import pyrin.security.session.services as session_services

from pyrin.database.decorators import session_factory
from pyrin.database.orm.query.base import CoreQuery
from pyrin.database.orm.session.base import CoreSession
from pyrin.database.session_factory.base import SessionFactoryBase


@session_factory()
class RequestScopedSessionFactory(SessionFactoryBase):
    """
    request scoped session factory class.
    """

    def __init__(self, **options):
        """
        initializes an instance of RequestScopedSessionFactory.
        """

        super().__init__(**options)

    def _create_session_factory(self, engine):
        """
        creates a database request scoped session factory and binds it to
        given engine and returns it.
        the scope is current request, so each request will get
        it's own session from start to end.

        :param Engine engine: database engine.

        :returns: database session
        :rtype: Session
        """

        session_configs = config_services.get_section('database', 'request_scoped_session')
        return scoped_session(sessionmaker(bind=engine, class_=CoreSession,
                                           query_cls=CoreQuery, **session_configs),
                              scopefunc=session_services.get_current_request)

    def is_request_bounded(self):
        """
        gets a value determining that this session factory
        type should be bounded into request.

        :rtype: bool
        """

        return True



# -*- coding: utf-8 -*-
"""
database migration alembic handlers show module.
"""

from pyrin.cli.interface import CLIHandlerOptionsMetadata
from pyrin.database.migration.alembic.decorators import alembic_cli_handler
from pyrin.database.migration.alembic.interface import AlembicCLIHandlerBase


@alembic_cli_handler()
class ShowCLIHandler(AlembicCLIHandlerBase):
    """
    show cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of ShowCLIHandler.
        """

        super().__init__('show')

    def _generate_custom_cli_handler_options_metadata(self):
        """
        generates custom cli handler options metadata.

        :rtype: list[CLIHandlerOptionsMetadata]
        """

        revision = CLIHandlerOptionsMetadata('revision', None)
        options = [revision]

        return options



# -*- coding: utf-8 -*-
"""
token handlers package.
"""

from pyrin.packaging.context import Package


class TokenHandlersPackage(Package):
    """
    token handlers package class.
    """

    NAME = __name__
    DEPENDS = []



# -*- coding: utf-8 -*-
"""
database migration alembic package.
"""

from pyrin.packaging.context import Package


class DatabaseMigrationAlembicPackage(Package):
    """
    database migration alembic package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'database.migration.alembic.component'
    ALEMBIC_CONFIG_STORE = 'alembic'



# -*- coding: utf-8 -*-
"""
serializer entity module.
"""

from pyrin.converters.serializer.base import SerializerBase
from pyrin.utils.sqlalchemy import entity_to_dict


class CoreEntitySerializer(SerializerBase):
    """
    core entity serializer class.
    """

    @classmethod
    def serialize(cls, value, **options):
        """
        serializes the given value.

        :param CoreEntity value: entity value to be serialized.

        :keyword bool exposed_only: if set to False, it returns all
                                    columns of the entity as dict.
                                    if not provided, defaults to True.

        :type: dict
        """

        exposed = options.get('exposed_only', True)
        return entity_to_dict(value, exposed_only=exposed)



# -*- coding: utf-8 -*-
"""
utils response module.
"""

from flask import make_response as flask_response

from pyrin.core.context import DTO
from pyrin.core.enumerations import ServerErrorResponseCodeEnum
from pyrin.settings.static import DEFAULT_STATUS_CODE


def make_response(message, **options):
    """
    makes a response from given inputs.

    :param str message: response message.

    :keyword int code: response code.
                       defaults to `DEFAULT_STATUS_CODE`, if not provided.

    :rtype: CoreResponse
    """

    code = options.get('code', DEFAULT_STATUS_CODE)
    response = DTO(code=code, message=message), code
    return flask_response(response)


def make_error_response(message, **options):
    """
    makes an error response from given inputs.

    :param str message: error message.

    :keyword int code: error code.
                       defaults to `INTERNAL_SERVER_ERROR` code, if not provided.

    :rtype: CoreResponse
    """

    code = options.get('code', ServerErrorResponseCodeEnum.INTERNAL_SERVER_ERROR)
    return make_response(message, code=code)


def make_exception_response(exception, **options):
    """
    makes an error response from given exception.
    if the exception does not have code, defaults to `INTERNAL_SERVER_ERROR` code.

    :param Exception exception: exception instance.

    :keyword int code: error code.
                       tries to get it from exception itself, if not provided.
                       defaults to `INTERNAL_SERVER_ERROR` code.
                       if not available in exception.

    :rtype: CoreResponse
    """

    message = getattr(exception, 'description', str(exception))
    code = options.get('code', None)
    if code is None:
        code = getattr(exception, 'code', ServerErrorResponseCodeEnum.INTERNAL_SERVER_ERROR)

    return make_error_response(message, code=code)



# -*- coding: utf-8 -*-
"""
database migration alembic handlers heads module.
"""

from pyrin.cli.interface import CLIHandlerOptionsMetadata
from pyrin.database.migration.alembic.decorators import alembic_cli_handler
from pyrin.database.migration.alembic.handlers.base import AlembicReportingCLIHandlerBase


@alembic_cli_handler()
class HeadsCLIHandler(AlembicReportingCLIHandlerBase):
    """
    heads cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of HeadsCLIHandler.
        """

        super().__init__('heads')

    def _generate_custom_cli_handler_options_metadata(self):
        """
        generates custom cli handler options metadata.

        :rtype: list[CLIHandlerOptionsMetadata]
        """

        resolve_dependencies = CLIHandlerOptionsMetadata('resolve_dependencies', None,
                                                         {True: '--resolve-dependencies',
                                                          False: None})
        options = [resolve_dependencies]

        return options



# -*- coding: utf-8 -*-
"""
cli decorators module.
"""

import inspect

from functools import update_wrapper

import pyrin.database.migration.alembic.services as alembic_services

from pyrin.utils.custom_print import print_error


def cli(func):
    """
    decorator to specify a method that handles a cli command.
    the method name will be used as the handler name for that cli handler.
    decorated method could modify inputs and return a dict of modified
    inputs if required.

    :param callable func: function.

    :returns: decorated function.
    :rtype: callable
    """

    def decorator(*args, **kwargs):
        """
        decorates the given function and executes the relevant cli handler.

        :param object args: function arguments.
        :param object kwargs: function keyword arguments.
        """

        try:
            original_inputs = inspect.getcallargs(func, *args, **kwargs)
            original_inputs.pop('self', None)
            original_inputs.pop('cls', None)
            modified_inputs = func(*args, **kwargs)
            original_inputs.update(**(modified_inputs or {}))
            alembic_services.execute(func.__name__, **original_inputs)

        except Exception as error:
            print_error(str(error), force=True)

    return update_wrapper(decorator, func)



# -*- coding: utf-8 -*-
"""
model manager module.
"""

from sqlalchemy.util._collections import AbstractKeyedTuple

from pyrin.core.context import Manager
from pyrin.database.model.base import CoreEntity


class ModelManager(Manager):
    """
    model manager class.
    """

    def is_core_entity(self, value):
        """
        gets a value indicating that input value is an instance of `CoreEntity`.
        this method could be used where direct reference to `CoreEntity`
        produces an error.

        :param object value: value to be checked.

        :rtype: bool
        """

        return isinstance(value, CoreEntity)

    def is_abstract_keyed_tuple(self, value):
        """
        gets a value indicating that input value is an instance
        of `AbstractKeyedTuple`. `AbstractKeyedTuple` objects are
        those objects that are returned by sqlalchemy `Query`
        with columns or multiple entities.

        :param object value: value to be checked.

        :rtype: bool
        """

        return isinstance(value, AbstractKeyedTuple)



# -*- coding: utf-8 -*-
"""
locale babel interface module.
"""

from pyrin.cli.interface import CLIHandlerBase, CLIHandlerOptionsMetadata


class BabelCLIHandlerBase(CLIHandlerBase):
    """
    babel cli handler base class.
    all babel cli handlers must be subclassed from this.
    """

    def __init__(self, name):
        """
        initializes an instance of BabelCLIHandlerBase.

        :param str name: the handler name that should be registered
                         with. this name must be the exact name that
                         this handler must emmit to cli.
        """

        super().__init__(name)

    def _inject_common_cli_options(self, commands):
        """
        injecting some common cli options into the given list.

        :param list commands: a list of all commands and their
                              values to be sent to cli command.
        """

        bounded_options = ['pybabel']
        for i in range(len(bounded_options)):
            commands.insert(i, bounded_options[i])

    def _generate_common_cli_handler_options_metadata(self):
        """
        generates common cli handler options metadata.

        :rtype: list[CLIHandlerOptionsMetadata]
        """

        help_option = CLIHandlerOptionsMetadata('help', None, {True: '--help', False: None})
        output_file = CLIHandlerOptionsMetadata('output_file', '--output-file')

        options = [help_option, output_file]

        return options



# -*- coding: utf-8 -*-
"""
locale package.
"""

from pyrin.packaging.context import Package


class LocalePackage(Package):
    """
    locale package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'globalization.locale.component'



# -*- coding: utf-8 -*-
"""
users package.
"""

from pyrin.packaging import Package


class UsersPackage(Package):
    """
    users package class.
    """

    NAME = __name__
    DEPENDS = []
    COMPONENT_NAME = 'security.users.component'



# -*- coding: utf-8 -*-
"""
database migration alembic handlers history module.
"""

from pyrin.cli.interface import CLIHandlerOptionsMetadata
from pyrin.database.migration.alembic.decorators import alembic_cli_handler
from pyrin.database.migration.alembic.handlers.base import AlembicReportingCLIHandlerBase


@alembic_cli_handler()
class HistoryCLIHandler(AlembicReportingCLIHandlerBase):
    """
    history cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of HistoryCLIHandler.
        """

        super().__init__('history')

    def _generate_custom_cli_handler_options_metadata(self):
        """
        generates custom cli handler options metadata.

        :rtype: list[CLIHandlerOptionsMetadata]
        """

        indicate_current = CLIHandlerOptionsMetadata('indicate_current', None,
                                                     {True: '--indicate-current',
                                                      False: None})
        revision_range = CLIHandlerOptionsMetadata('revision_range', '--rev-range')

        options = [indicate_current, revision_range]

        return options



# -*- coding: utf-8 -*-
"""
database migration alembic handlers revision module.
"""

from pyrin.cli.interface import CLIHandlerOptionsMetadata
from pyrin.database.migration.alembic.decorators import alembic_cli_handler
from pyrin.database.migration.alembic.interface import AlembicCLIHandlerBase


@alembic_cli_handler()
class RevisionCLIHandler(AlembicCLIHandlerBase):
    """
    revision cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of RevisionCLIHandler.
        """

        super().__init__('revision')

    def _generate_custom_cli_handler_options_metadata(self):
        """
        generates custom cli handler options metadata.

        :rtype: list[CLIHandlerOptionsMetadata]
        """

        message = CLIHandlerOptionsMetadata('message', '--message')
        autogenerate = CLIHandlerOptionsMetadata('autogenerate', None,
                                                 {True: '--autogenerate', False: None})
        sql = CLIHandlerOptionsMetadata('sql', None, {True: '--sql', False: None})
        head = CLIHandlerOptionsMetadata('head', '--head')
        splice = CLIHandlerOptionsMetadata('splice', None, {True: '--splice', False: None})
        branch_label = CLIHandlerOptionsMetadata('branch_label', '--branch-label')
        version_path = CLIHandlerOptionsMetadata('version_path', '--version-path')
        revision_id = CLIHandlerOptionsMetadata('revision_id', '--rev-id')
        depends_on = CLIHandlerOptionsMetadata('depends_on', '--depends-on')

        options = [message, autogenerate, sql,
                   head, splice, branch_label,
                   version_path, revision_id, depends_on]

        return options



# -*- coding: utf-8 -*-
"""
database orm sql operators mixin module.
"""

from datetime import datetime

from sqlalchemy.sql.operators import ColumnOperators

import pyrin.utils.datetime as datetime_utils

from pyrin.utils.sqlalchemy import like_prefix, like_exact_prefix, \
    like_suffix, like_exact_suffix


class ColumnOperatorsMixin(ColumnOperators):
    """
    column operators mixin.
    this class provides some practical benefits
    to its subclasses for sql operations.
    """

    def between(self, cleft, cright, symmetric=False, **options):
        """
        produces an `expression.between` clause against the parent
        object, given the lower and upper range. this method is overridden
        to be able to handle datetime values more practical.

        :param object cleft: lower bound of clause.
        :param object cright: upper bound of clause.

        :param bool symmetric: specifies to emmit `between symmetric` to database.
                               note that not all databases support symmetric.
                               but `between symmetric` is equivalent to
                               `between least(a, b) and greatest(a, b)`.

        :keyword bool consider_begin_of_day: specifies that consider begin
                                             of day for lower datetime.
                                             defaults to True if not provided.
                                             this only has effect on datetime value.

        :keyword bool consider_end_of_day: specifies that consider end
                                           of day for upper datetime.
                                           defaults to True if not provided.
                                           this only has effect on datetime value.

        :returns: operation result
        """

        consider_begin_of_day = options.get('consider_begin_of_day', True)
        consider_end_of_day = options.get('consider_end_of_day', True)
        is_lower_datetime = isinstance(cleft, datetime)
        is_upper_datetime = isinstance(cright, datetime)
        if consider_begin_of_day is True and is_lower_datetime is True:
            cleft = datetime_utils.begin_of_day(cleft)

        if consider_end_of_day is True and is_upper_datetime is True:
            cright = datetime_utils.end_of_day(cright)

        # swapping values in case of user mistake.
        if symmetric is True and is_lower_datetime is True \
                and is_upper_datetime is True and cleft > cright:
            cleft, cright = cright, cleft
            symmetric = False

        return super().between(cleft, cright, symmetric)

    def _process_like_prefix(self, value, **options):
        """
        processes the value that should be prefixed to
        value for `like` expression based on given options.

        :param str value: expression to be compared.

        :keyword bool exact: specifies that value should be emitted without any
                             modifications. defaults to False if not provided.

        :keyword int begin_count: count of `_` chars to be attached to beginning.
                                  if not provided, `%` will be used.

        :note begin_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                           if the provided value goes upper than this limit,
                           a `%` will be attached instead of it. this limit
                           is for security reason.

        :returns: operation result
        """

        exact = options.get('exact', False)
        if exact is False:
            begin_wrapper = like_prefix, [value]
            begin_count = options.get('begin_count', None)
            if begin_count is not None:
                begin_wrapper = like_exact_prefix, [value, begin_count]

            value = begin_wrapper[0](*begin_wrapper[1])

        return value

    def _process_like_suffix(self, value, **options):
        """
        processes the value that should be suffixed to
        value for `like` expression based on given options.

        :param str value: expression to be compared.

        :keyword bool exact: specifies that value should be emitted without any
                             modifications. defaults to False if not provided.

        :keyword int end_count: count of `_` chars to be attached to end.
                                if not provided, `%` will be used.

        :note end_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                         if the provided value goes upper than this limit,
                         a `%` will be attached instead of it. this limit
                         is for security reason.

        :returns: operation result
        """

        exact = options.get('exact', False)
        if exact is False:
            end_wrapper = like_suffix, [value]
            end_count = options.get('end_count', None)
            if end_count is not None:
                end_wrapper = like_exact_suffix, [value, end_count]

            value = end_wrapper[0](*end_wrapper[1])

        return value

    def like(self, other, escape=None, **options):
        """
        implements the `like` operator.
        this method is overridden to apply `%` or couple of `_`
        place holders on both sides of input string in default mode.
        if you want to apply place holder to just one side, use
        `startswith()` or `endswith()` methods.

        :param str other: expression to be compared.

        :param str escape: optional escape character,
                           renders the `escape` keyword.

        :keyword bool exact: specifies that value should be emitted without any
                             modifications. defaults to False if not provided.

        :keyword int begin_count: count of `_` chars to be attached to beginning.
                                  if not provided, `%` will be used.

        :keyword int end_count: count of `_` chars to be attached to end.
                                if not provided, `%` will be used.

        :note begin_count, end_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                                      if the provided value goes upper than this limit,
                                      a `%` will be attached instead of it. this limit
                                      is for security reason.

        :returns: operation result
        """

        other = self._process_like_prefix(other, **options)
        other = self._process_like_suffix(other, **options)
        return super().like(other, escape)

    def ilike(self, other, escape=None, **options):
        """
        implements the `ilike` operator.
        this method is overridden to apply `%` or couple of `_`
        place holders on both sides of input string in default mode.
        if you want to apply place holder to just one side, use
        `istartswith()` or `iendswith()` methods.

        :param str other: expression to be compared.

        :param str escape: optional escape character,
                           renders the `escape` keyword.

        :keyword bool exact: specifies that value should be emitted without any
                             modifications. defaults to False if not provided.

        :keyword int begin_count: count of `_` chars to be attached to beginning.
                                  if not provided, `%` will be used.

        :keyword int end_count: count of `_` chars to be attached to end.
                                if not provided, `%` will be used.

        :note begin_count, end_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                                      if the provided value goes upper than this limit,
                                      a `%` will be attached instead of it. this limit
                                      is for security reason.

        :returns: operation result
        """

        other = self._process_like_prefix(other, **options)
        other = self._process_like_suffix(other, **options)
        return super().like(other, escape)

    def istartswith(self, other, **kwargs):
        """
        implements the `startswith` operator.
        produces an `ilike` expression that tests against a match for the
        start of a string value. for example:
        column like <other> || '%' or column like <other> || '_'
        this method provides a case-insensitive variant of `startswith()`
        method.

        :param str other: expression to be compared.

        :keyword bool autoescape: establishes an escape character within the like
                                  expression, then applies it to all occurrences of
                                  `%`, `_` and the escape character itself within the
                                  comparison value.

        :keyword str escape: optional escape character,
                             renders the `escape` keyword.

        :keyword bool exact: specifies that value should be emitted without any
                             modifications. defaults to False if not provided.

        :keyword int end_count: count of `_` chars to be attached to end.
                                if not provided, `%` will be used.

        :note end_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                         if the provided value goes upper than this limit,
                         a `%` will be attached instead of it. this limit
                         is for security reason.

        :returns: operation result
        """



# -*- coding: utf-8 -*-
"""
database hooks module.
"""

import pyrin.database.services as database_services

from pyrin.core.context import Hook
from pyrin.packaging.decorators import packaging_hook
from pyrin.packaging.hooks import PackagingHookBase


class DatabaseHookBase(Hook):
    """
    database hook base class.
    """

    def __init__(self):
        """
        initializes an instance of DatabaseHookBase.
        """

        super().__init__()

    def after_session_factories_configured(self):
        """
        this method will be called after all database
        session factories have been configured.
        """
        pass


@packaging_hook()
class PackagingHook(PackagingHookBase):
    """
    packaging hook class.
    """

    def __init__(self):
        """
        initializes an instance of PackagingHook.
        """

        super().__init__()

    def after_packages_loaded(self):
        """
        this method will be called after all application packages has been loaded.
        """

        # we have to configure session factories after all models have
        # been loaded to enable multiple database connections if needed.
        database_services.configure_session_factories()



# -*- coding: utf-8 -*-
"""
database migration alembic handlers stamp module.
"""

from pyrin.cli.interface import CLIHandlerOptionsMetadata
from pyrin.database.migration.alembic.decorators import alembic_cli_handler
from pyrin.database.migration.alembic.interface import AlembicCLIHandlerBase


@alembic_cli_handler()
class StampCLIHandler(AlembicCLIHandlerBase):
    """
    stamp cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of StampCLIHandler.
        """

        super().__init__('stamp')

    def _generate_custom_cli_handler_options_metadata(self):
        """
        generates custom cli handler options metadata.

        :rtype: list[CLIHandlerOptionsMetadata]
        """

        sql = CLIHandlerOptionsMetadata('sql', None, {True: '--sql', False: None})
        tag = CLIHandlerOptionsMetadata('tag', '--tag')
        purge = CLIHandlerOptionsMetadata('purge', None, {True: '--purge', False: None})
        revisions = CLIHandlerOptionsMetadata('revisions', None)

        options = [sql, tag, purge, revisions]

        return options



# -*- coding: utf-8 -*-
"""
locale babel component module.
"""

from pyrin.application.decorators import component
from pyrin.application.context import Component
from pyrin.globalization.locale.babel import LocaleBabelPackage
from pyrin.globalization.locale.babel.manager import LocaleBabelManager


@component(LocaleBabelPackage.COMPONENT_NAME)
class LocaleBabelComponent(Component, LocaleBabelManager):
    """
    locale babel component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
packaging exceptions module.
"""

from pyrin.core.exceptions import CoreException


class PackagingManagerException(CoreException):
    """
    packaging manager exception.
    """
    pass


class InvalidPackageNameError(PackagingManagerException):
    """
    invalid package name error.
    """
    pass


class ComponentModuleNotFoundError(PackagingManagerException):
    """
    component module not found error.
    """
    pass



# -*- coding: utf-8 -*-
"""
database orm query base module.
"""

import inspect

from sqlalchemy import inspection, log, func
from sqlalchemy.orm import Query, lazyload
from sqlalchemy.orm.attributes import InstrumentedAttribute

from pyrin.core.globals import LIST_TYPES, _
from pyrin.database.model.base import CoreEntity
from pyrin.database.orm.sql.schema.base import CoreColumn
from pyrin.database.services import get_current_store
from pyrin.database.orm.query.exceptions import ColumnsOutOfScopeError, \
    UnsupportedQueryStyleError


@inspection._self_inspects
@log.class_logger
class CoreQuery(Query):
    """
    core query class.
    this is the application default query class.
    it extends sqlalchemy `Query` class.
    """

    def __init__(self, entities, session=None, **options):
        """
        initializes an instance of CoreQuery.

        :param tuple entities: entities or columns that are needed for query.
        :param Session session: optional session object to bind this query to it.

        :keyword Union[type, tuple[type]] scope: class type of the entities that this
                                                 query instance will work on. if the
                                                 query is working on multiple entities,
                                                 this value must be a tuple of all class
                                                 types of that entities.

                                                 for example: if you set
                                                 `entities=SomeEntity.id, AnotherEntity.name`
                                                 you should leave `scope=None` to skip validation
                                                 or you could set
                                                 `scope=(SomeEntity, AnotherEntity)`
                                                 this way validation succeeds, but if
                                                 you set `scope=SomeEntity`
                                                 then the query will not be executed
                                                 and an error would be raised.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        scope = options.get('scope', None)
        if scope is not None:
            self._prepare_validation(entities, scope)

        super().__init__(entities, session)

    def _validate_scope(self, entities, scope):
        """
        validates the query for given entities based on provided scope.

        :param tuple entities: entities or columns that needed for query.

        :param Union[type, tuple[type]] scope: class type of the entities that this
                                               query instance will work on. if the
                                               query is working on multiple entities,
                                               this value must be a tuple of all class
                                               types of that entities.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        scope_classes = set(entity for entity in scope if inspect.isclass(entity)
                            and issubclass(entity, CoreEntity))

        requested_classes = set(entity for entity in entities if inspect.isclass(entity)
                                and issubclass(entity, CoreEntity))

        requested_classes_by_column = set(column.class_ for column in entities if
                                          isinstance(column, InstrumentedAttribute)
                                          and issubclass(getattr(column, 'class_', type),
                                                         CoreEntity))

        all_requested_classes = requested_classes.union(requested_classes_by_column)

        if not all_requested_classes.issubset(scope_classes):
            raise ColumnsOutOfScopeError(_('Requested columns are out of scope of this '
                                           'query. Please revise the requested columns.'))

    def _prepare_validation(self, entities, scope):
        """
        prepares entities and scope and validates them if required.

        :param tuple entities: entities or columns that needed for query.

        :param Union[type, tuple[type]] scope: class type of the entities that this
                                               query instance will work on. if the
                                               query is working on multiple entities,
                                               this value must be a tuple of all class
                                               types of that entities.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        if scope is None or entities is None:
            return

        if not isinstance(scope, LIST_TYPES):
            scope = (scope,)

        if not isinstance(entities, LIST_TYPES):
            entities = (entities,)

        if len(scope) <= 0 or len(entities) <= 0:
            return

        self._validate_scope(entities, scope)

    def count(self, **options):
        """
        returns the count of rows the sql formed by this `Query` would return.
        this method is overridden to prevent inefficient count() of sqlalchemy `Query`
        which produces a subquery.

        this method generates a single sql query like below:
        select count(column, ...)
        from table
        where ...

        :keyword bool distinct: specifies that count should
                                be executed on distinct select.
                                defaults to False if not provided.

        :keyword bool fallback: specifies that count should
                                be executed using original sqlalchemy
                                function which produces a subquery.
                                defaults to False if not provided.

        :rtype: int
        """

        fallback = options.get('fallback', False)
        needs_fallback = False
        columns = []
        for single_column in self.selectable.columns:
            if not isinstance(single_column, CoreColumn):
                if fallback is False:
                    raise UnsupportedQueryStyleError('Current query does not have columns '
                                                     'of type [{column_type}] in its expression. '
                                                     'if you need to apply a "DISTINCT" keyword, '
                                                     'you should apply it by passing '
                                                     '"distinct=True" keyword to count() method '
                                                     'and do not apply it in query structure '
                                                     'itself. for example instead of writing '
                                                     '"store.query(distinct(Entity.id)).count()" '
                                                     'you should write this in the following form '
                                                     '"store.query(Entity.id).count(distinct=True)'
                                                     '". but if you want the sqlalchemy original '
                                                     'style of count() which produces a subquery, '
                                                     'it is also possible to fallback to that '
                                                     'default sqlalchemy count() but keep in '
                                                     'mind that, that method is not efficient. '
                                                     'you could pass "fallback=True" in options '
                                                     'to fallback to default mode if overridden '
                                                     'count() method failed to provide count.'
                                                     .format(column_type=CoreColumn))
                else:
                    needs_fallback = True
                    break

            fullname = single_column.fullname()
            if fullname not in (None, ''):
                columns.append(fullname)

        if needs_fallback is True:
            return super().count()

        func_count = func.count()
        if len(columns) > 0:
            distinct = options.get('distinct', False)
            column_clause = ', '.join(columns)
            if distinct is True:
                column_clause = 'distinct {clause}'.format(clause=column_clause)
            func_count = func.count(column_clause)

        statement = self.options(lazyload('*')).statement.with_only_columns(
            [func_count]).order_by(None)

        store = get_current_store()
        return store.execute(statement).scalar()



# -*- coding: utf-8 -*-
"""
authorization manager module.
"""

import pyrin.security.services as security_services
import pyrin.security.users.services as user_services
import pyrin.security.session.services as session_services

from pyrin.core.globals import _, LIST_TYPES
from pyrin.core.context import Manager
from pyrin.security.authorization.exceptions import AuthorizationFailedError, \
    UserNotAuthenticatedError, UserIsNotActiveError


class AuthorizationManager(Manager):
    """
    authorization manager class.
    this class is intended to provide some services needed in pyrin application.
    the top level application must extend this class considering business requirements.
    """

    def authorize(self, user, permissions, **options):
        """
        authorizes the given user for specified permissions.
        if user does not have each one of specified permissions,
        an error will be raised.

        :param user: user identity to authorize permissions for.

        :param Union[PermissionBase, list[PermissionBase]] permissions: permissions to check
                                                                        for user authorization.

        :raises UserNotAuthenticatedError: user not authenticated error.
        :raises UserIsNotActiveError: user is not active error.
        :raises AuthorizationFailedError: authorization failed error.
        """

        if user is None:
            raise UserNotAuthenticatedError(_('User has not been authenticated.'))

        if not user_services.is_active(user, **options):
            message = _('User [{user}] is not active.')
            raise UserIsNotActiveError(message.format(user=str(user)))

        # we must check whether input permissions object is iterable.
        # if not, we make it manually.
        required_permissions = permissions
        if not isinstance(permissions, LIST_TYPES):
            required_permissions = [permissions]

        if security_services.has_permission(user, required_permissions, **options) is not True:
            message = _('User [{user}] does not have required permission(s) {permissions}.')
            raise AuthorizationFailedError(message.format(user=str(user),
                                                          permissions=str(required_permissions)))

    def is_authorized(self, permissions, user=None, **options):
        """
        gets a value indicating that specified user is authorized for given permissions.

        :param Union[PermissionBase, list[PermissionBase]] permissions: permissions to check
                                                                        for authorization.

        :param user: user identity to be checked for authorization.
                     if not provided, current user will be used.

        :rtype: bool
        """

        if user is None:
            user = session_services.get_current_user()

        try:
            self.authorize(user, permissions)
            return True
        except AuthorizationFailedError:
            return False



# -*- coding: utf-8 -*-
"""
caching handlers base module.
"""

import time

from threading import Lock
from copy import deepcopy

import pyrin.configuration.services as config_services

from pyrin.core.context import CoreObject, DTO
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.core.globals import NULL
from pyrin.utils.singleton import MultiSingletonMeta


class CacheItem(CoreObject):
    """
    cache item class.
    """

    def __init__(self, key, value, timeout):
        """
        initializes an instance of CacheItem.

        :param object key: hashable key for this cache item.
        :param value: value of the hash item.
        :param int timeout: timeout for this cache item before
                            invalidation in milliseconds.
        """

        super().__init__()

        self.key = key
        self.value = deepcopy(value)
        self.create_time = time.time()
        self.hit_count = 0
        self.miss_count = 0
        self.timeout = timeout

    def __str__(self):
        return 'CacheItem: [{key}-{value}]'.format(key=self.key, value=self.value)

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(self.key)

    def is_expired(self):
        """
        gets a value indicating that this cache item is expired.

        :rtype: bool
        """

        return (time.time() * 1000) - self.timeout > self.create_time * 1000


class CachingHandlerSingletonMeta(MultiSingletonMeta):
    """
    caching handler singleton meta class.
    this is a thread-safe implementation of singleton.
    """

    # a dictionary containing an instance of each type.
    # in the form of: {type: instance}
    _instances = dict()
    _lock = Lock()


class CachingHandlerBase(CoreObject, metaclass=CachingHandlerSingletonMeta):
    """
    caching handler base class.
    """

    def __init__(self, name, **options):
        """
        initializes an instance of CachingHandlerBase.

        :param str name: caching handler name.
                         this name will be used to get handlers.

        :keyword int timeout: timeout for items of this caching handler
                              before invalidation in milliseconds.
                              defaults to `caching.config` file `default_timeout`
                              value if not provided.
        """

        super().__init__()

        self._set_name(name)
        self._timeout = options.get('timeout', config_services.get('caching',
                                                                   'general',
                                                                   'default_timeout'))

    def __setitem__(self, key, value):
        return self.set(key, value)

    def __getitem__(self, key):
        return self.get(key)

    def __contains__(self, key):
        return self.contains(key)

    def set(self, key, value, timeout=None):
        """
        sets a new value into cached items.

        :param object key: hashable key of the cache to be registered.
        :param object value: value to be cached.

        :param int timeout: timeout for this value. defaults
                            to value from `caching.config`
                            if not provided.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()

    def get(self, key):
        """
        gets the value from cache.

        :param object key: key to get its value from cache.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: object
        """

        raise CoreNotImplementedError()

    def contains(self, key):
        """
        gets a value indicating that given key is existed in the cached items.

        :param object key: key to be checked for existence.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()

    def get_timeout(self):
        """
        gets timeout value for this handler items in milliseconds.

        :rtype: int
        """

        return self._timeout

    def remove(self, key):
        """
        removes the key from cached items.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError()


class DictCachingHandlerBase(CachingHandlerBase):
    """
    dict caching handler base class.
    """

    def __init__(self, name, **options):
        """
        initializes an instance of DictCachingHandlerBase.

        :param str name: caching handler name.
                         this name will be used to get handlers.

        :keyword int max_length: max length of this caching handler.
                                 it means how many elements this handler
                                 could cache before it starts to remove old items.
                                 defaults to `caching.config` file `default_max_length`
                                 value if not provided.

        :keyword int timeout: timeout for items of this caching handler
                              before invalidation in milliseconds.
                              defaults to `caching.config` file `default_timeout`
                              value if not provided.
        """

        super().__init__(name, **options)

        self._max_length = options.get('max_length', config_services.get('caching',
                                                                         'general',
                                                                         'default_max_length'))
        self._items = DTO()
        self._cache_lock = Lock()

    def __len__(self):
        return self.get_current_length()

    def get_max_length(self):
        """
        gets the max length of this caching handler.

        :rtype: int
        """

        return self._max_length

    def get_current_length(self):
        """
        gets the length of current items available in this caching handler.

        :rtype: int
        """

        return len(self._items)

    def clear(self):
        """
        removes all cached items.
        """

        self._items.clear()

    def set(self, key, value, timeout=None):
        """
        sets a new value into cached items.

        :param object key: hashable key of the cache to be registered.
        :param object value: value to be cached.

        :param int timeout: timeout for this value. defaults
                            to value from `caching.config`
                            if not provided.
        """

        item = CacheItem(key, value, timeout or self.get_timeout())
        try:
            self._cache_lock.acquire()
            self._set(item)
        finally:
            self._cache_lock.release()

    def _set(self, item):
        """
        puts the given item into cache.

        :param CacheItem item: item to be cached.
        """

        if self.get_current_length() > self.get_max_length():
            pass

        self._items[item.key] = item

    def get(self, key):
        """
        gets the value from cache.
        if the key does not existed, it returns `NULL` object.

        :param object key: key to get its value from cache.

        :rtype: object
        """

        item = self._items.get(key, NULL)
        if item is not NULL:
            self.remove(key)
            if item.is_expired():
                return NULL
            self._set(item)
            return item.value

        return item

    def contains(self, key):
        """
        gets a value indicating that given key is existed in the cached items.

        :param object key: key to be checked for existence.

        :rtype: bool
        """

        return key in self._items

    def remove(self, key):
        """
        removes the key from cached items.

        :param object key: key to get its value from cache.
        """

        del self._items[key]



# -*- coding: utf-8 -*-
"""
permission manager module.
"""

from pyrin.core.context import Context, Manager
from pyrin.security.permission.base import PermissionBase
from pyrin.security.permission.exceptions import InvalidPermissionTypeError, \
    DuplicatedPermissionError


class PermissionManager(Manager):
    """
    permission manager class.
    this class is intended to be an interface for top level application's permission
    package, so most methods of this class will raise CoreNotImplementedError.
    """

    def __init__(self):
        """
        initializes an instance of PermissionManager.
        """

        super().__init__()

        # holds a dict of all application's loaded permissions.
        # in the form of dict(str permission_id: PermissionBase permission)
        self.__permissions = Context()

    def register_permission(self, instance, **options):
        """
        registers the given permission.

        :param PermissionBase instance: permission instance to be registered.

        :raises InvalidPermissionTypeError: invalid permission type error.
        :raises DuplicatedPermissionError: duplicated permission error.
        """

        if not isinstance(instance, PermissionBase):
            raise InvalidPermissionTypeError('Input parameter [{instance}] is '
                                             'not an instance of [{base}].'
                                             .format(instance=instance,
                                                     base=PermissionBase))

        if instance.get_id() in self.__permissions:
            raise DuplicatedPermissionError('Permission [{permission}] has been '
                                            'already registered.'
                                            .format(permission=instance))

        self.__permissions[instance.get_id()] = instance

    def get_permissions(self, **options):
        """
        gets all registered permissions.

        :rtype: list[PermissionBase]
        """

        return self.__permissions.values()

    def synchronize_all(self, **options):
        """
        synchronizes all permissions with database.
        it creates or updates the available permissions.
        if you don't need permissions in your application
        leave this method not implemented.
        """
        pass



# -*- coding: utf-8 -*-
"""
deserializer dictionary module.
"""

import re

import flask.json as flask_json

import pyrin.converters.deserializer.services as deserializer_services

from pyrin.converters.deserializer.decorators import deserializer
from pyrin.core.context import DTO
from pyrin.core.globals import NULL
from pyrin.converters.deserializer.handlers.base import DeserializerBase, \
    StringPatternDeserializerBase


@deserializer()
class DictionaryDeserializer(DeserializerBase):
    """
    dictionary deserializer class.
    """

    def __init__(self, **options):
        """
        creates an instance of DictionaryDeserializer.
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes every possible value available in input dictionary.
        and gets a new deserialized dictionary, leaving the input unchanged.

        :param dict value: value that should be deserialized.

        :rtype: dict
        """

        if not self.is_deserializable(value, **options):
            return NULL

        result = DTO(**value)

        for key in result.keys():
            item = result.get(key)
            deserialized_value = NULL

            if self.is_deserializable(item, **options):
                deserialized_value = self.deserialize(item)
            else:
                deserialized_value = deserializer_services.deserialize(item, **options)

            if deserialized_value is not NULL:
                result[key] = deserialized_value
            continue

        return result

    def get_accepted_type(self):
        """
        gets the accepted type for this deserializer
        which could deserialize values from this type.

        :rtype: type
        """

        return dict


@deserializer()
class StringDictionaryDeserializer(StringPatternDeserializerBase):
    """
    string dictionary deserializer class.
    """

    # default min for this deserializer is 2 because
    # it should at least has { and } at both ends.
    DEFAULT_MIN = 2

    # matches a dictionary inside string, all of these values will be matched.
    # example: {}, {22: 1}, {'key1': 1, 'key2': 2}
    DICT_REGEX = re.compile(r'^{.*}$')

    def __init__(self, **options):
        """
        creates an instance of StringDictionaryDeserializer.

        :keyword list[tuple(Pattern, int)] accepted_formats: a list of custom accepted patterns
                                                             and their length for dictionary
                                                             deserialization.

        :type accepted_formats: list[tuple(Pattern format, int length)]
        """

        super().__init__(**options)

    def _deserialize(self, value, **options):
        """
        deserializes the given value.
        returns `NULL` object if deserialization fails.

        :param str value: value to be deserialized.

        :rtype: tuple
        """

        deserializable, pattern = self.is_deserializable(value, **options)
        if not deserializable:
            return NULL

        try:
            dict_value = flask_json.loads(value)
            if dict_value is not None:
                deserialized_value = deserializer_services.deserialize(dict_value, **options)
                return deserialized_value

            return NULL
        except Exception:
            return NULL

    def get_default_formats(self):
        """
        gets default accepted patterns that this
        deserializer could deserialize value from.

        :return: list(tuple(Pattern format, int length))

        :rtype: list(tuple(Pattern, int))
        """

        return [(self.DICT_REGEX, self.UNDEF_LENGTH)]



# -*- coding: utf-8 -*-
"""
cli exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class CLIManagerException(CoreException):
    """
    cli manager exception.
    """
    pass


class CLIManagerBusinessException(CoreBusinessException, CLIManagerException):
    """
    cli manager business exception.
    """
    pass


class MetaDataOptionsParamNameIsRequiredError(CLIManagerException):
    """
    metadata options param name is required error.
    """
    pass


class ParamValueIsNotMappedToCLIError(CLIManagerBusinessException):
    """
    param value is not mapped to cli error.
    """
    pass


class InvalidCLIHandlerNameError(CLIManagerException):
    """
    invalid cli handler name error.
    """
    pass


class InvalidCLIHandlerTypeError(CLIManagerException):
    """
    invalid cli handler type error.
    """
    pass


class DuplicatedCLIHandlerError(CLIManagerException):
    """
    duplicated cli handler error.
    """
    pass


class CLIHandlerNotFoundError(CLIManagerBusinessException):
    """
    cli handler not found error.
    """
    pass


class InvalidOptionsMetaDataTypeError(CLIManagerException):
    """
    invalid options metadata type error.
    """
    pass



# -*- coding: utf-8 -*-
"""
logging adapters module.
"""

from logging import LoggerAdapter

import pyrin.security.session.services as session_services

from pyrin.core.context import DTO
from pyrin.core.exceptions import CoreNotImplementedError


class BaseLoggerAdapter(LoggerAdapter):
    """
    base logger adapter class.
    all logger adapters should be subclassed from this.
    """

    def __init__(self, logger):
        """
        initializes an instance of BaseLoggerAdapter.

        :param Logger logger: logger instance to be wrapped.
        """

        super().__init__(logger, DTO())

        # these attributes have been added for compatibility
        # with loggers common api.
        self.handlers = logger.handlers
        self.level = logger.level
        self.propagate = logger.propagate
        self.parent = logger.parent

    def addHandler(self, hdlr):
        """
        adds the specified handler to this logger.
        this method has been added for compatibility
        with loggers common api.

        :param Handler hdlr: logger handler to be added.
        """

        self.logger.addHandler(hdlr)

    def removeHandler(self, hdlr):
        """
        removes the specified handler from this logger.
        this method has been added for compatibility
        with loggers common api.

        :param Handler hdlr: logger handler to be removed.
        """

        self.logger.removeHandler(hdlr)

    def process(self, msg, kwargs):
        """
        processes the logging message and keyword arguments passed in to
        a logging call to insert contextual information. you can either
        manipulate the message itself, the keyword args or both. return
        the message and kwargs modified (or not) to suit your needs.

        normally, you'll only need to override this one method in a
        `LoggerAdapter` subclass for your specific needs.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: tuple(str message, dict kwargs)

        :rtype: tuple
        """

        raise CoreNotImplementedError()


class RequestInfoLoggerAdapter(BaseLoggerAdapter):
    """
    request info logger adapter.
    this adapter adds request info into generated logs.
    """

    def process(self, msg, kwargs):
        """
        processes the logging message and keyword arguments passed in to
        a logging call to insert contextual information. you can either
        manipulate the message itself, the keyword args or both. return
        the message and kwargs modified (or not) to suit your needs.

        :returns: tuple(str message, dict kwargs)

        :rtype: tuple
        """

        client_request = session_services.get_safe_current_request()
        custom_message = '[{client_request}]: {message}'.format(client_request=client_request,
                                                                message=msg)
        return custom_message, kwargs



# -*- coding: utf-8 -*-
"""
application base module.
"""

import os.path

from time import time

import dotenv

from flask import Flask, request
from flask.app import setupmethod
from flask.ctx import has_request_context

import pyrin.packaging.services as packaging_services
import pyrin.configuration.services as config_services
import pyrin.security.authentication.services as authentication_services
import pyrin.security.session.services as session_services
import pyrin.logging.services as logging_services
import pyrin.database.model.services as model_services
import pyrin.utils.misc as misc_utils
import pyrin.utils.path as path_utils

from pyrin.application.container import _set_app
from pyrin.api.router.handlers.protected import ProtectedRoute
from pyrin.application.enumerations import ApplicationStatusEnum
from pyrin.application.hooks import ApplicationHookBase
from pyrin.application.mixin import SignalMixin
from pyrin.converters.json.decoder import CoreJSONDecoder
from pyrin.converters.json.encoder import CoreJSONEncoder
from pyrin.converters.serializer.entity import CoreEntitySerializer
from pyrin.converters.serializer.keyed_tuple import CoreKeyedTupleSerializer
from pyrin.core.context import DTO, Manager
from pyrin.core.globals import LIST_TYPES
from pyrin.core.mixin import HookMixin
from pyrin.packaging import PackagingPackage
from pyrin.packaging.component import PackagingComponent
from pyrin.application.context import CoreResponse, CoreRequest, ApplicationContext, \
    ApplicationComponent, ApplicationSingletonMeta
from pyrin.application.context import Component
from pyrin.settings.static import DEFAULT_COMPONENT_KEY
from pyrin.utils.custom_print import print_warning
from pyrin.utils.dictionary import make_key_upper
from pyrin.application.exceptions import DuplicateContextKeyError, InvalidComponentTypeError, \
    InvalidComponentIDError, DuplicateComponentIDError, DuplicateRouteURLError, \
    InvalidRouteFactoryTypeError, InvalidApplicationStatusError, \
    ApplicationInScriptingModeError, ComponentAttributeError, \
    ApplicationIsNotSubclassedError


class Application(Flask, HookMixin, SignalMixin,
                  metaclass=ApplicationSingletonMeta):
    """
    application class.
    server must initialize an instance of a subclass of this class at startup.
    """

    # the application looks for these stores to configure itself.
    # they should be present in settings folder of the upper
    # level application.
    CONFIG_STORES = ['application',
                     'communication',
                     'environment']

    # pyrin default settings path will be registered in application context with this key.
    PYRIN_DEFAULT_SETTINGS_CONTEXT_KEY = 'default_settings_path'

    # settings path will be registered in application context with this key.
    SETTINGS_CONTEXT_KEY = 'settings_path'

    # migrations path will be registered in application context with this key.
    MIGRATIONS_CONTEXT_KEY = 'migrations_path'

    # locale path will be registered in application context with this key.
    LOCALE_CONTEXT_KEY = 'locale_path'

    # pyrin main package path will be registered in application context with this key.
    PYRIN_PATH_CONTEXT_KEY = 'pyrin_path'

    # application main package path will be registered in application context with this key.
    APPLICATION_PATH_CONTEXT_KEY = 'application_path'

    # pyrin root path will be registered in application context with this key.
    # pyrin root path is where pyrin main package is located.
    ROOT_PYRIN_PATH_CONTEXT_KEY = 'root_pyrin_path'

    # application root path will be registered in application context with this key.
    # root path is where application main package and other files are located.
    ROOT_APPLICATION_PATH_CONTEXT_KEY = 'root_application_path'

    # default packaging component to be used by application.
    # if you want to change the default one, you could subclass
    # Application and set `packaging_component_class` to your desired one.
    # this design is compatible with other flask subclassing features.
    # note that your custom packaging component class should not use `@component`
    # decorator to register itself, application will register it instead.
    packaging_component_class = PackagingComponent

    url_rule_class = ProtectedRoute
    response_class = CoreResponse
    request_class = CoreRequest
    json_decoder = CoreJSONDecoder
    json_encoder = CoreJSONEncoder
    entity_serializer = CoreEntitySerializer
    keyed_tuple_serializer = CoreKeyedTupleSerializer
    _hook_type = ApplicationHookBase

    def __init__(self, **options):
        """
        initializes an instance of Application.

        :keyword bool host_matching: set `url_map.host_matching` attribute.
                                     defaults to False.

        :keyword str subdomain_matching: consider the subdomain relative to
                                         `SERVER_NAME` when matching routes.
                                         defaults to False.

        :keyword str template_folder: the folder that contains the templates that should
                                      be used by the application. defaults to
                                      `templates` folder in the root path of the application.

        :keyword str instance_path: an alternative instance path for the application.
                                    by default the folder `instance` next to the
                                    package or module is assumed to be the instance path.

        :keyword bool instance_relative_config: if set to `True` relative filenames
                                                for loading the config are assumed to
                                                be relative to the instance path instead
                                                of the application root.

        :keyword str root_path: flask by default will automatically calculate the path
                                to the root of the application. in certain situations
                                this cannot be achieved (for instance if the package
                                is a python 3 namespace package) and needs to be
                                manually defined.

        :keyword bool scripting_mode: specifies that the application has been started in
                                      scripting mode. some application hooks will not
                                      get fired when the app runs in scripting mode.
                                      defaults to False, if not provided.

        :keyword str settings_directory: settings directory name.
                                         if not provided, defaults to `settings`.

        :keyword str migrations_directory: migrations directory name.
                                           if not provided, defaults to `migrations`.

        :keyword str locale_directory: locale directory name.
                                       if not provided, defaults to `locale`.
        """

        self._assert_is_subclassed()

        self.__status = ApplicationStatusEnum.INITIALIZING
        self._scripting_mode = options.get('scripting_mode', False)

        flask_kw = self._remove_flask_unrecognized_keywords(**options)
        # we should pass `static_folder=None` to prevent flask from
        # adding static route on startup, then we register required static routes
        # through a different mechanism later.
        super().__init__(self.get_application_name(), static_folder=None, **flask_kw)

        # Flask does not call 'super()' in its '__init__()' method.
        # so we have to start initialization of other parents manually.
        HookMixin.__init__(self, **options)

        self._context = ApplicationContext()
        self._components = ApplicationComponent()

        # we should register some packages manually because they are referenced
        # in `application.base` module and could not be loaded automatically
        # because packaging package will not handle them.
        self._register_required_components()

        # setting the application instance in application container module.
        _set_app(self)

        # we should load application at this stage to be able to perform pytest
        # tests after application has been fully loaded. because if we call
        # application.run(), we could not continue execution of other codes.
        self._load(**options)
        self._set_status(ApplicationStatusEnum.RUNNING)

    def _remove_flask_unrecognized_keywords(self, **options):
        """
        removes any keyword argument which is not recognized
        by `Flask.__init__()` method. it returns a new dict
        with all recognized keyword arguments.

        :rtype: dict
        """

        result = dict(**options)
        result.pop('scripting_mode', None)
        result.pop('settings_directory', None)
        result.pop('migrations_directory', None)
        result.pop('locale_directory', None)

        return result

    def _assert_is_subclassed(self):
        """
        asserts that current application instance is subclassed from
        `Application` class and is not a direct instance of `Application` itself.

        :raises ApplicationIsNotSubclassedError: application is not subclassed error.
        """

        if type(self) == Application:
            raise ApplicationIsNotSubclassedError('Current application instance is a direct '
                                                  'instance of "Application". you must subclass '
                                                  'from "Application" in your project and create '
                                                  'an instance of that class to run your '
                                                  'application. this is needed for pyrin to be '
                                                  'able to resolve different paths correctly.')

    def is_scripting_mode(self):
        """
        gets a value indicating that application has been started in scripting mode.
        some application hooks will not fire in this mode. like 'before_application_start'.
        """

        return self._scripting_mode

    @setupmethod
    def _register_required_components(self):
        """
        registers required components that the Application needs to
        reference to them immediately.
        this type of components could not be registered using @component decorator,
        because they will be referenced before Application instance gets initialized.

        note that implementation-wise, application package should depend on other
        packages services as few as possible, so be careful if you needed some external
        packages services inside application package, probably it would be better to
        change your design that enforces the application to use other packages services.
        """

        self.register_component(self.packaging_component_class(
            PackagingPackage.COMPONENT_NAME))

    def _set_status(self, status):
        """
        sets the application status.
        status must be from ApplicationStatusEnum.

        :param str status: application status.

        :note status:
            INITIALIZING = 'Initializing'
            LOADING = 'Loading'
            RUNNING = 'Running'
            TERMINATED = 'Terminated'

        raises InvalidApplicationStatusError: invalid application status error.
        """

        if status not in ApplicationStatusEnum:
            raise InvalidApplicationStatusError('Application status [{state}] is not valid.'
                                                .format(state=status))

        old_status = self.__status
        self.__status = status

        self._application_status_changed(old_status, status)

    def get_status(self):
        """
        gets the application status.

        :rtype: str
        """

        return self.__status

    def add_context(self, key, value, **options):
        """
        adds the given key and it's value into the application context.

        :param str key: related key for storing application context.
        :param object value: related value for storing in application context.

        :keyword bool replace: specifies that if there is already a value with
                               the same key in application context, it should be updated
                               with new value, otherwise raise an error. defaults to False.

        :raises DuplicateContextKeyError: duplicate context key error.
        """

        replace = options.get('replace', False)
        if replace is not True and key in self._context:
            raise DuplicateContextKeyError('Key [{key}] is already available in application '
                                           'context and "replace=True" option is not set, so '
                                           'the new value could not be added.'
                                           .format(key=key))

        self._context[key] = value

    def get_context(self, key):
        """
        gets the application context value that belongs to given key.

        :param str key: key for requested application context.

        :returns: related value to given key.
        """

        return self._context[key]

    @setupmethod
    def register_component(self, component, **options):
        """
        registers given application component or replaces the existing one
        if `replace=True` is provided. otherwise, it raises an error
        on adding an instance which it's id is already available
        in registered components.

        :param Component component: component instance.

        :keyword bool replace: specifies that if there is another registered
                               component with the same id, replace it with the new one.
                               otherwise raise an error. defaults to False.

        :raises InvalidComponentTypeError: invalid component type error.
        :raises InvalidComponentIDError: invalid component id error.
        :raises DuplicateComponentIDError: duplicate component id error.
        """

        if not isinstance(component, Component):
            raise InvalidComponentTypeError('Input parameter [{component}] is not '
                                            'an instance of [{instance}].'
                                            .format(component=component,
                                                    instance=Component))

        if not isinstance(component, Manager):
            raise InvalidComponentTypeError('Input parameter [{component}] is not '
                                            'an instance of [{instance}]. each component '
                                            'class must be subclassed from its respective '
                                            'manager class of the same package and that '
                                            'manager class must be subclassed from [{instance}].'
                                            .format(component=component,
                                                    instance=Manager))

        if not isinstance(component.get_id(), tuple) or \
                len(component.get_id()[0].strip()) == 0:
            raise InvalidComponentIDError('Component [{component}] does '
                                          'not have a valid component id.'
                                          .format(component=component))

        # checking whether is there any registered component with the same id.
        if component.get_id() in self._components.keys():
            replace = options.get('replace', False)

            if replace is not True:
                raise DuplicateComponentIDError('There is another registered component with '
                                                'id [{id}] but "replace" option is not set, so '
                                                'component [{instance}] could not be registered.'
                                                .format(id=component.get_id(),
                                                        instance=component))

            old_instance = self._components[component.get_id()]

            # we should update all list and dict attributes and also those that have
            # three consecutive underscores in their names of new component with values
            # from old_instance to prevent loss of any attribute value (for example values
            # that has been added using decorators).
            # this has an obvious caveat, and it is that child classes could not do
            # any customizations on these attributes in their `__init__` method.
            component = self._set_component_attributes(old_instance, component)

            print_warning('Component [{old_instance}] is going to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance, new_instance=component))

        self._components[component.get_id()] = component

    def remove_component(self, component_id):
        """
        removes application component with given id.

        :param tuple component_id: component id to be removed.
        :type component_id: tuple(str, object)

        :raises ComponentAttributeError: component attribute error.
        """

        if component_id not in self._components:
            raise ComponentAttributeError('Component [{component_id}] is not '
                                          'available in application components.'
                                          .format(component_id=component_id))

        self._components.pop(component_id)

    def _get_safe_current_request(self):
        """
        gets current request in a safe manner.
        meaning that if there is no request in current context,
        returns None instead of raising an error.

        :rtype: CoreRequest
        """

        if has_request_context() is True:
            with request:
                return request

        return None

    def get_component(self, component_name, **options):
        """
        gets the specified application component.

        :param str component_name: component name.

        :keyword object component_custom_key: component custom key.
                                              if not provided, tries to get it from
                                              request object, if not found,
                                              `DEFAULT_COMPONENT_KEY` will be used.

        :raises InvalidComponentNameError: invalid component name error.

        :rtype: Component
        """

        component_custom_key = options.get('component_custom_key', None)

        if component_custom_key is None:
            component_custom_key = self._extract_component_custom_key()

        # checking whether is there any custom implementation.
        component_custom_id = \
            Component.make_component_id(component_name,
                                        component_custom_key=component_custom_key)

        if component_custom_id in self._components.keys():
            return self._components[component_custom_id]

        # getting default component.
        component_default_id = Component.make_component_id(component_name)
        return self._components[component_default_id]

    def _extract_component_custom_key(self):
        """
        gets `component_custom_key` from current request.
        note that if application is in any state other than `RUNNING`,
        it always returns the default component key.

        :rtype: object
        """

        # before application reaches the `RUNNING` state,
        # we should always use the default component key.
        if self.get_status() != ApplicationStatusEnum.RUNNING:
            return DEFAULT_COMPONENT_KEY

        # this method is the only place that we should access request
        # directly and not from session services, because of circular calling problem.
        client_request = self._get_safe_current_request()
        if client_request is not None:
            return client_request.component_custom_key

        return DEFAULT_COMPONENT_KEY

    def _load(self, **options):
        """
        loads application configs and components.
        """

        self._set_status(ApplicationStatusEnum.LOADING)
        self._resolve_all_paths(**options)
        self._load_environment_variables()

        packaging_services.load_components(**options)

        # we should call this method after loading components
        # to be able to use configuration package.
        self._load_configs(**options)

        # calling `after_application_loaded` method of all registered hooks.
        self._after_application_loaded()

        # calling `before_application_start` method of all registered hooks.
        # this hook should not be called when the app is in scripting mode.
        if self.is_scripting_mode() is False:
            self._before_application_start()

    def _resolve_all_paths(self, **options):
        """
        resolves all required paths for application.
        """

        self._resolve_pyrin_main_package_path(**options)
        self._resolve_application_main_package_path(**options)
        self._resolve_pyrin_root_path(**options)
        self._resolve_application_root_path(**options)
        self._resolve_default_settings_path()
        self._resolve_settings_path(**options)
        self._resolve_migrations_path(**options)
        self._resolve_locale_path(**options)

    def _load_configs(self, **options):
        """
        loads all configurations related to application package.
        """

        config_services.load_configurations(*self.CONFIG_STORES, **options)
        for store_name in self.CONFIG_STORES:
            config_dict = config_services.get_all(store_name, **options)
            self.configure(config_dict)

    def _configure(self, config_store):
        """
        configures the application with given dict.
        all keys will be converted to uppercase for flask compatibility.

        :param dict config_store: a dictionary containing configuration key/values.
        """

        upper_dict = make_key_upper(config_store)
        self.config.update(upper_dict)

    def configure(self, config_store):
        """
        configures the application with given dict.
        all keys will be converted to uppercase for flask compatibility.

        :param dict config_store: a dictionary containing configuration key/values.
        """

        self._configure(config_store)

    def run(self, host=None, port=None, debug=None,
            load_dotenv=True, **options):
        """
        runs the Application instance.

        :param str host: the hostname to listen on. set this to `0.0.0.0` to
                         have the server available externally as well. defaults to
                         `127.0.0.1` or the host in the `SERVER_NAME`
                         config variable if present.

        :param int port: the port of the webserver. defaults to `5000` or the
                         port defined in the `SERVER_NAME` config variable if present.

        :param bool debug: if given, enable or disable debug mode.

        :param bool load_dotenv: load the nearest `.env` and `.flaskenv`
                                 files to set environment variables. will also
                                 change the working directory to the directory
                                 containing the first file found.

        :keyword bool use_reloader: specifies that the server should automatically
                                    restart the python process if modules were changed.

        :keyword bool use_debugger: specifies that the werkzeug debugging
                                    system should be used.

        :keyword bool threaded: specifies that the process should handle
                                each request in a separate thread.

        :raises ApplicationInScriptingModeError: application in scripting mode error.
        """

        if self.is_scripting_mode() is True:
            raise ApplicationInScriptingModeError('Application has been initialized in '
                                                  'scripting mode, so it could not be run.')

        super().run(host, port, debug, load_dotenv, **options)

    def dispatch_request(self):
        """
        does the request dispatching. matches the URL and returns the
        return value of the view or error handlers. this does not have to
        be a response object. in order to convert the return value to a
        proper response object, call `make_response` function.

        :raises AuthenticationFailedError: authentication failed error.
        """

        # we have to override whole `dispatch_request` method to be able to customize it,
        # because of the flask design that everything is embedded inside
        # the `dispatch_request` method.

        client_request = session_services.get_current_request()
        if client_request.routing_exception is not None:
            self.raise_routing_exception(client_request)

        route = client_request.url_rule

        # if we provide automatic options for this URL and the
        # request came with the OPTIONS method, reply automatically.
        if getattr(route, 'provide_automatic_options', False) \
           and client_request.method == 'OPTIONS':
            return self.make_default_options_response()

        # otherwise call the handler for this route.
        return route.handle(client_request.get_inputs())

    def _authenticate(self, client_request):
        """
        authenticates given request.

        :param CoreRequest client_request: request to be authenticated.

        :raises AuthenticationFailedError: authentication failed error.
        """

        authentication_services.authenticate(client_request)

    def make_response(self, rv):
        """
        converts the return value from a view function
        to an instance of `CoreResponse`.

        note that the `rv` value before passing to base method must
        be a `tuple`, `dict`, `str` or `CoreResponse`. otherwise
        the value will not be json serialized and it causes an error.

        :param object rv: the return value from the view function.

        :rtype: CoreResponse
        """

        rv = self._convert_result(rv)

        if rv is None:
            rv = DTO()

        # we could not return a list as response, so we wrap the
        # result in a dict when we want to return a list.
        if isinstance(rv, list):
            rv = DTO(items=rv)

        # we should wrap all single values into a dict before returning it to client.
        if not isinstance(rv, (tuple, dict, CoreResponse)):
            rv = DTO(value=rv)

        return super().make_response(rv)

    def _convert_result(self, rv):
        """
        converts the return value if needed.
        this method could be overridden in subclasses.

        :param object rv: the return value from the view function.

        :rtype: object
        """

        if isinstance(rv, list) and len(rv) > 0:
            if model_services.is_abstract_keyed_tuple(rv[0]):
                rv = self.keyed_tuple_serializer.serialize_list(rv)
            elif model_services.is_core_entity(rv[0]):
                rv = self.entity_serializer.serialize_list(rv)

        elif model_services.is_abstract_keyed_tuple(rv):
            rv = self.keyed_tuple_serializer.serialize(rv)

        elif model_services.is_core_entity(rv):
            rv = self.entity_serializer.serialize(rv)

        return rv

    @setupmethod
    def add_url_rule(self, rule, endpoint=None, view_func=None,
                     provide_automatic_options=None, **options):
        """
        connects a url rule. if a view_func is provided it will be registered with the endpoint.
        if there is another rule with the same url and `replace=True` option is provided,
        it will be replaced, otherwise an error will be raised.

        :param str rule: the url rule as string.

        :param str endpoint: the endpoint for the registered url rule.
                             pyrin itself assumes the url rule as endpoint.

        :param callable view_func: the function to call when serving a request to the
                                   provided endpoint.

        :param bool provide_automatic_options: controls whether the `OPTIONS` method should be
                                               added automatically.
                                               this can also be controlled by setting the
                                               `view_func.provide_automatic_options = False`
                                               before adding the rule.

        :keyword tuple(str) methods: http methods that this rule should handle.
                                     if not provided, defaults to `GET`.

        :keyword tuple(PermissionBase) permissions: tuple of all required permissions
                                                    to access this route's resource.

        :keyword bool login_required: specifies that this route could not be accessed
                                      if the requester has not a valid token.
                                      defaults to True if not provided.

        :keyword bool replace: specifies that this route must replace
                               any existing route with the same url or raise
                               an error if not provided. defaults to False.

        :raises DuplicateRouteURLError: duplicate route url error.
        """

        methods = options.get('methods', ())
        if not isinstance(methods, LIST_TYPES):
            options.update(methods=(methods,))

        replace = options.get('replace', False)

        # setting endpoint to url rule instead of view function name,
        # to be able to have the same function names on different url rules.
        if endpoint is None:
            endpoint = rule

        # checking whether is there any registered route with the same url.
        old_rule = None
        for rule_item in self.url_map.iter_rules(endpoint=None):
            if rule_item.rule == rule:
                old_rule = rule_item
                break

        if old_rule is not None:
            if replace is True:
                self.url_map._rules.remove(old_rule)
                if old_rule.endpoint in self.view_functions.keys():
                    self.view_functions.pop(old_rule.endpoint)
                if old_rule.endpoint in self.url_map._rules_by_endpoint.keys():
                    self.url_map._rules_by_endpoint.pop(old_rule.endpoint)

                print_warning('Registered route for url [{url}] is '
                              'going to be replaced by a new route.'
                              .format(url=rule))
            else:
                raise DuplicateRouteURLError('There is another registered route with the '
                                             'same url [{url}], but "replace" option is not '
                                             'set, so the new route could not be registered.'
                                             .format(url=rule))

        # we have to put `view_function=view_func` into options to be able to deliver it
        # to route initialization in the super method. that's because of flask design
        # that does not forward all params to inner method calls. and this is the less
        # ugly way in comparison to overriding the whole `add_url_rule` function.
        options.update(view_function=view_func)

        super().add_url_rule(rule, endpoint, view_func,
                             provide_automatic_options, **options)

    @setupmethod
    def register_route_factory(self, factory):
        """
        registers a route factory as application url rule class.

        :param callable factory: route factory.
                                 it could be a class or a factory method.

        :raises InvalidRouteFactoryTypeError: invalid route factory type error.
        """

        if not callable(factory):
            raise InvalidRouteFactoryTypeError('Input parameter [{factory}] is not callable.'
                                               .format(factory=str(factory)))

        print_warning('Registered route factory [{old_factory}] is '
                      'going to be replaced by a new route factory [{new_factory}].'
                      .format(old_factory=str(self.url_rule_class),
                              new_factory=str(factory)))

        self.url_rule_class = factory

    def get_current_route_factory(self):
        """
        gets current route factory in use.

        :rtype: callable
        """

        return self.url_rule_class

    def get_default_settings_path(self):
        """
        gets the pyrin default settings path.

        :rtype: str
        """

        return self.get_context(self.PYRIN_DEFAULT_SETTINGS_CONTEXT_KEY)

    def get_settings_path(self):
        """
        gets the application settings path.

        :rtype: str
        """

        return self.get_context(self.SETTINGS_CONTEXT_KEY)

    def get_migrations_path(self):
        """
        gets the application migrations path.

        :rtype: str
        """

        return self.get_context(self.MIGRATIONS_CONTEXT_KEY)

    def get_locale_path(self):
        """
        gets the application locale path.

        :rtype: str
        """

        return self.get_context(self.LOCALE_CONTEXT_KEY)

    def get_application_main_package_path(self):
        """
        gets the application main package path.

        :rtype: str
        """

        return self.get_context(self.APPLICATION_PATH_CONTEXT_KEY)

    def get_pyrin_root_path(self):
        """
        gets pyrin root path in which pyrin package is located.

        :rtype: str
        """

        return self.get_context(self.ROOT_PYRIN_PATH_CONTEXT_KEY)

    def get_application_root_path(self):
        """
        gets the application root path in which application package is located.

        :rtype: str
        """

        return self.get_context(self.ROOT_APPLICATION_PATH_CONTEXT_KEY)

    def get_pyrin_main_package_path(self):
        """
        gets pyrin main package path.

        :rtype: str
        """

        return self.get_context(self.PYRIN_PATH_CONTEXT_KEY)

    def get_configs(self):
        """
        gets a shallow copy of application's configuration dictionary.

        :rtype: dict
        """

        return self.config.copy()

    def _resolve_settings_path(self, **options):
        """
        resolves the application settings path. the resolved path will
        be accessible by `SETTINGS_CONTEXT_KEY` inside application context.

        :keyword str settings_directory: settings directory name.
                                         if not provided, defaults to `settings`.
        """

        main_package_path = self.get_application_main_package_path()
        settings_path = os.path.join(main_package_path, options.get('settings_directory',
                                                                    'settings'))
        settings_path = os.path.abspath(settings_path)

        if not os.path.isdir(settings_path):
            print_warning('Application settings path [{path}] does not exist. '
                          'pyrin default settings from [{default}] will be used. '
                          'you could copy any of default setting files into '
                          '[{path}] and change values on your preference. '
                          'DO NOT use pyrin default settings in production!'
                          .format(path=settings_path,
                                  default=self.get_default_settings_path()))

            settings_path = self.get_default_settings_path()

        self.add_context(self.SETTINGS_CONTEXT_KEY, settings_path)

    def _resolve_default_settings_path(self):
        """
        resolves the pyrin default settings path. the resolved path will
        be accessible by `PYRIN_DEFAULT_SETTINGS_CONTEXT_KEY` inside
        application context.
        """

        pyrin_main_package_path = self.get_pyrin_main_package_path()
        settings_path = os.path.join(pyrin_main_package_path, 'settings', 'default')
        settings_path = os.path.abspath(settings_path)
        self.add_context(self.PYRIN_DEFAULT_SETTINGS_CONTEXT_KEY, settings_path)

    def _resolve_migrations_path(self, **options):
        """
        resolves the application migrations path. the resolved path will
        be accessible by `MIGRATIONS_CONTEXT_KEY` inside application context.

        :keyword str migrations_directory: migrations directory name.
                                           if not provided, defaults to `migrations`.
        """

        main_package_path = self.get_application_main_package_path()
        migrations_path = os.path.join(main_package_path, options.get('migrations_directory',
                                                                      'migrations'))
        migrations_path = os.path.abspath(migrations_path)

        self.add_context(self.MIGRATIONS_CONTEXT_KEY, migrations_path)

    def _resolve_locale_path(self, **options):
        """
        resolves the application locale path. the resolved path will
        be accessible by `LOCALE_CONTEXT_KEY` inside application context.

        :keyword str locale_directory: locale directory name.
                                       if not provided, defaults to `locale`.
        """

        main_package_path = self.get_application_main_package_path()
        locale_path = os.path.join(main_package_path, options.get('locale_directory',
                                                                  'locale'))
        locale_path = os.path.abspath(locale_path)

        self.add_context(self.LOCALE_CONTEXT_KEY, locale_path)

    def _resolve_application_main_package_path(self, **options):
        """
        resolves the application main package path and registers it
        in application context with `APPLICATION_PATH_CONTEXT_KEY` key.
        """

        main_package_path = path_utils.get_main_package_path(self.__module__)
        self.add_context(self.APPLICATION_PATH_CONTEXT_KEY, main_package_path)

    def _resolve_pyrin_main_package_path(self, **options):
        """
        resolves pyrin main package path and registers it
        in application context with `PYRIN_PATH_CONTEXT_KEY` key.
        """

        pyrin_main_package = path_utils.get_pyrin_main_package_path()
        self.add_context(self.PYRIN_PATH_CONTEXT_KEY, pyrin_main_package)

    def _resolve_pyrin_root_path(self, **options):
        """
        resolves pyrin root path and registers it in application
        context with `ROOT_PYRIN_PATH_CONTEXT_KEY` key.
        """

        main_package_path = self.get_pyrin_main_package_path()
        root_path = os.path.join(main_package_path, '..')
        root_path = os.path.abspath(root_path)
        self.add_context(self.ROOT_PYRIN_PATH_CONTEXT_KEY, root_path)

    def _resolve_application_root_path(self, **options):
        """
        resolves application root path and registers it in application
        context with `ROOT_APPLICATION_PATH_CONTEXT_KEY` key.
        """

        main_package_path = self.get_application_main_package_path()
        root_path = os.path.join(main_package_path, '..')
        root_path = os.path.abspath(root_path)
        self.add_context(self.ROOT_APPLICATION_PATH_CONTEXT_KEY, root_path)

    def _load_environment_variables(self):
        """
        loads all environment variables defined in a `.env` file in application
        root path. if the file does not exist, it will be ignored.
        """

        root_path = self.get_application_root_path()
        env_file = os.path.join(root_path, '.env')

        if not os.path.isfile(env_file):
            print_warning('Could not find ".env" file in application root path [{root_path}].'
                          .format(root_path=root_path))
            return

        dotenv.load_dotenv(env_file)

    def process_response(self, response):
        """
        this method is overridden to add required attributes into response object.

        :param CoreResponse response: response object.

        :rtype: CoreResponse
        """

        client_request = session_services.get_current_request()
        response.request_date = client_request.request_date
        response.request_id = client_request.request_id
        response.user = session_services.get_current_user()

        return super().process_response(response)

    def full_dispatch_request(self):
        """
        dispatches the request and on top of that performs request pre and
        postprocessing as well as http exception catching and error handling.
        this method has been overridden to log before and after request dispatching.
        """

        client_request = session_services.get_current_request()
        try:
            self._authenticate(client_request)

        except Exception as error:
            logging_services.exception(str(error))

        process_start_time = time()
        logging_services.info('request received with params: [{params}]'
                              .format(params=client_request.get_inputs(silent=True)))

        response = super().full_dispatch_request()

        process_end_time = time()
        logging_services.info('request executed in [{time} ms].'
                              .format(time='{:0.3f}'
                                      .format((process_end_time - process_start_time) * 1000)))

        logging_services.debug('[{response}] returned with result: [{result}]'
                               .format(response=response,
                                       result=response.get_data(as_text=True)))

        return response

    def _set_component_attributes(self, old_instance, new_instance):
        """
        replaces all list and dict attributes from old instance into new instance.
        all the attributes which their name starts with three underscores will also
        be replaced from old instance into new instance.

        :param Component old_instance: old component instance to get attributes from.
        :param Component new_instance: new component instance to set its attributes.

        :rtype: Component
        """

        if old_instance is None or new_instance is None:
            return new_instance

        all_attributes = vars(old_instance)
        required_attributes = DTO()
        for attribute_name in all_attributes.keys():
            if isinstance(all_attributes[attribute_name], (list, dict)) \
                    or '___' in attribute_name:
                required_attributes[attribute_name] = all_attributes[attribute_name]

        return misc_utils.set_attributes(new_instance, **required_attributes)

    def _after_application_loaded(self):
        """
        this method will call `after_application_loaded` method of all registered hooks.
        """

        for hook in self._get_hooks():
            hook.after_application_loaded()

    def _before_application_start(self):
        """
        this method will call `before_application_start` method of all registered hooks.
        """

        for hook in self._get_hooks():
            hook.before_application_start()

    def get_application_name(self):
        """
        gets the application name.

        :rtype: str
        """

        return path_utils.get_main_package_name(self.__module__)

    def _prepare_termination(self, signal_number):
        """
        prepares for termination.

        :param int signal_number: signal number that caused termination.
        """

        self._set_status(ApplicationStatusEnum.TERMINATED)

    def _application_status_changed(self, old_status, new_status):
        """
        this method will call `application_status_changed` method of all registered hooks.

        :param str old_status: old application status.
        :param str new_status: new application status.

        :note status:
            INITIALIZING = 'Initializing'
            LOADING = 'Loading'
            RUNNING = 'Running'
            TERMINATED = 'Terminated'
        """

        for hook in self._get_hooks():
            hook.application_status_changed(old_status, new_status)



# -*- coding: utf-8 -*-
"""
router handlers exceptions module.
"""

from pyrin.core.enumerations import ClientErrorResponseCodeEnum
from pyrin.core.exceptions import CoreException, CoreBusinessException
from pyrin.security.authentication.exceptions import AuthenticationFailedError


class RouterHandlerException(CoreException):
    """
    router handler exception.
    """
    pass


class RouterHandlerBusinessException(CoreBusinessException,
                                     RouterHandlerException):
    """
    router handler business exception.
    """
    pass


class InvalidViewFunctionTypeError(RouterHandlerException):
    """
    invalid view function type error.
    """
    pass


class MaxContentLengthLimitMismatchError(RouterHandlerException):
    """
    max content length limit mismatch error.
    """
    pass


class FreshTokenRequiredError(AuthenticationFailedError,
                              RouterHandlerBusinessException):
    """
    fresh token required error.
    """
    pass


class LargeContentError(RouterHandlerBusinessException):
    """
    large content error.
    """

    def __init__(self, *args, **kwargs):
        super(LargeContentError, self).__init__(*args, **kwargs)
        self.code = ClientErrorResponseCodeEnum.PAYLOAD_TOO_LARGE


class PermissionTypeError(RouterHandlerException):
    """
    permission type error.
    """
    pass



# -*- coding: utf-8 -*-
"""
application services module.
"""

from pyrin.application.container import _get_app


def add_context(key, value, **options):
    """
    adds the given key and it's value into the application context.

    :param str key: related key for storing application context.
    :param object value: related value for storing in application context.

    :keyword bool replace: specifies that if there is already a value with
                           the same key in application context, it should be updated
                           with new value, otherwise raise an error. defaults to False.

    :raises DuplicateContextKeyError: duplicate context key error.
    """

    get_current_app().add_context(key, value, **options)


def get_context(key):
    """
    gets the application context value that belongs to given key.

    :param str key: key for requested application context.

    :returns: related value to given key.
    """

    return get_current_app().get_context(key)


def register_component(component, **options):
    """
    registers given application component or replaces the existing one
    if `replace=True` is provided. otherwise, it raises an error
    on adding an instance which it's id is already available
    in registered components.

    :param Component component: component instance.

    :keyword bool replace: specifies that if there is another registered
                           component with the same id, replace it with the new one.
                           otherwise raise an error. defaults to False.

    :raises InvalidComponentTypeError: invalid component type error.
    :raises InvalidComponentIDError: invalid component id error.
    :raises DuplicateComponentIDError: duplicate component id error.
    """

    get_current_app().register_component(component, **options)


def remove_component(component_id):
    """
    removes application component with given id.

    :param tuple component_id: component id to be removed.
    :type component_id: tuple(str, object)

    :raises ComponentAttributeError: component attribute error.
    """

    get_current_app().remove_component(component_id)


def get_component(component_name, **options):
    """
    gets the specified application component.

    :param str component_name: component name.

    :keyword object component_custom_key: component custom key.
                                          if not provided, tries to get it from
                                          request object, if not found,
                                          `DEFAULT_COMPONENT_KEY` will be used.

    :rtype: Component
    """

    return get_current_app().get_component(component_name, **options)


def register_error_handler(code_or_exception, func):
    """
    registers the given function as an error handler for given code or exception type.

    :param Union[int, Exception] code_or_exception: code or exception to
                                                    register error handler for.

    :param callable func: function to register it as an error handler.
    """

    get_current_app().register_error_handler(code_or_exception, func)


def register_before_request_handler(func):
    """
    registers the given function into application before request handlers.

    :param callable func: function to register it into before request handlers.
    """

    get_current_app().before_request(func)


def register_after_request_handler(func):
    """
    registers the given function into application after request handlers.

    :param callable func: function to register it into after request handlers.
    """

    get_current_app().after_request(func)


def register_teardown_request_handler(func):
    """
    registers the given function into application teardown request handlers.
    teardown request handlers should not return any value
    and also should not raise any exception.

    :param callable func: function to register it into teardown request handlers.
    """

    get_current_app().teardown_request(func)


def add_url_rule(rule, endpoint=None, view_func=None,
                 provide_automatic_options=None, **options):
    """
    connects a url rule. if a view_func is provided it will be registered with the endpoint.
    if there is another rule with the same url and `replace=True` option is provided,
    it will be replaced, otherwise an error will be raised.

    :param str rule: the url rule as string.

    :param str endpoint: the endpoint for the registered url rule.
                         pyrin itself assumes the url rule as endpoint.

    :param callable view_func: the function to call when serving a request to the
                               provided endpoint.

    :param bool provide_automatic_options: controls whether the `OPTIONS` method should be
                                           added automatically.
                                           this can also be controlled by setting the
                                           `view_func.provide_automatic_options = False`
                                           before adding the rule.

    :keyword tuple(str) methods: http methods that this rule should handle.
                                 if not provided, defaults to `GET`.

    :keyword tuple(PermissionBase) permissions: tuple of all required permissions
                                                to access this route's resource.

    :keyword bool login_required: specifies that this route could not be accessed
                                  if the requester has not a valid token.
                                  defaults to True if not provided.

    :keyword bool replace: specifies that this route must replace
                           any existing route with the same url or raise
                           an error if not provided. defaults to False.

    :raises DuplicateRouteURLError: duplicate route url error.
    """

    get_current_app().add_url_rule(rule, endpoint, view_func,
                                   provide_automatic_options, **options)


def register_route_factory(factory):
    """
    registers a route factory as application url rule class.

    :param callable factory: route factory.
                             it could be a class or a factory method.

    :raises InvalidRouteFactoryTypeError: invalid route factory type error.
    """

    get_current_app().register_route_factory(factory)


def get_current_route_factory():
    """
    gets current route factory in use.

    :rtype: callable
    """

    return get_current_app().get_current_route_factory()


def get_default_settings_path():
    """
    gets the pyrin default settings path.

    :rtype: str
    """

    return get_current_app().get_default_settings_path()


def get_settings_path():
    """
    gets the application settings path.

    :rtype: str
    """

    return get_current_app().get_settings_path()


def get_migrations_path():
    """
    gets the application migrations path.

    :rtype: str
    """

    return get_current_app().get_migrations_path()


def get_locale_path():
    """
    gets the application locale path.

    :rtype: str
    """

    return get_current_app().get_locale_path()


def get_application_main_package_path():
    """
    gets the application main package path.

    :rtype: str
    """

    return get_current_app().get_application_main_package_path()


def get_pyrin_root_path():
    """
    gets pyrin root path in which pyrin package is located.

    :rtype: str
    """

    return get_current_app().get_pyrin_root_path()


def get_application_root_path():
    """
    gets the application root path in which application package is located.

    :rtype: str
    """

    return get_current_app().get_application_root_path()


def get_pyrin_main_package_path():
    """
    gets pyrin main package path.

    :rtype: str
    """

    return get_current_app().get_pyrin_main_package_path()


def get_configs():
    """
    gets a shallow copy of application's configuration dictionary.

    :rtype: dict
    """

    return get_current_app().get_configs()


def configure(config_store):
    """
    configures the application with given dict.
    all keys will be converted to uppercase for flask compatibility.

    :param dict config_store: a dictionary containing configuration key/values.
    """

    get_current_app().configure(config_store)


def get_current_app():
    """
    gets the instance of current running application.

    :rtype: Application
    """

    return _get_app()


def register_hook(instance):
    """
    registers the given instance into application hooks.

    :param ApplicationHookBase instance: application hook instance to be registered.

    :raises InvalidApplicationHookTypeError: invalid application hook type error.
    """

    get_current_app().register_hook(instance)


def is_scripting_mode():
    """
    gets a value indicating that application has been started in scripting mode.
    some application hooks will not fire in this mode. like 'before_application_start'.
    """

    return get_current_app().is_scripting_mode()


def get_application_name():
    """
    gets the application name.

    :rtype: str
    """

    return get_current_app().get_application_name()



# -*- coding: utf-8 -*-
"""
utils path module.
"""

import os
import sys

from pyrin.utils.environment import is_windows


PATH_SEPARATOR = '/' if not is_windows() else '\\'


def get_module_file_path(module_name):
    """
    gets the absolute file path of module with given name.

    :param str module_name: module name to get its file path.

    :rtype: str
    """

    return os.path.abspath(sys.modules[module_name].__file__)


def get_main_package_name(module_name):
    """
    gets the main package name from given module name.
    for example for `pyrin.database.manager` module, it
    returns `pyrin` as the main package name.

    :param str module_name: module name to get its root package name.

    :rtype str
    """

    return module_name.split('.')[0]


def get_main_package_path(module_name):
    """
    gets the absolute path of the main package of module with given name.

    :param str module_name: module name to get its main package path.

    :rtype: str
    """

    relative_module_path = module_name.replace('.', PATH_SEPARATOR)
    root_package = get_main_package_name(module_name)
    absolute_module_path = get_module_file_path(module_name)
    temp_absolute_module_path = absolute_module_path.replace(relative_module_path, '*')
    excess_part = temp_absolute_module_path.split('*')[-1]
    list_path = list(temp_absolute_module_path)

    for i in range(-1, -len(list_path), -1):
        if list_path[i] == '*':
            list_path[i] = root_package
            break

    main_package_path = ''.join(list_path)
    main_package_path = main_package_path.replace('*', relative_module_path)
    main_package_path = main_package_path.replace(excess_part, '').rstrip(PATH_SEPARATOR)

    return main_package_path


def get_pyrin_main_package_name():
    """
    gets the name of pyrin main package name.
    it would always be `pyrin` in normal cases.

    :rtype: str
    """

    return get_main_package_name(__name__)


def get_pyrin_main_package_path():
    """
    gets the absolute path of pyrin main package.

    :rtype: str
    """

    return get_main_package_path(__name__)



# -*- coding: utf-8 -*-
"""
database migration alembic handlers merge module.
"""

from pyrin.cli.interface import CLIHandlerOptionsMetadata
from pyrin.database.migration.alembic.decorators import alembic_cli_handler
from pyrin.database.migration.alembic.interface import AlembicCLIHandlerBase


@alembic_cli_handler()
class MergeCLIHandler(AlembicCLIHandlerBase):
    """
    merge cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of MergeCLIHandler.
        """

        super().__init__('merge')

    def _generate_custom_cli_handler_options_metadata(self):
        """
        generates custom cli handler options metadata.

        :rtype: list[CLIHandlerOptionsMetadata]
        """

        message = CLIHandlerOptionsMetadata('message', '--message')
        branch_label = CLIHandlerOptionsMetadata('branch_label', '--branch-label')
        revision_id = CLIHandlerOptionsMetadata('revision_id', '--rev-id')
        revisions = CLIHandlerOptionsMetadata('revisions', None)

        options = [message, branch_label, revision_id, revisions]

        return options



# -*- coding: utf-8 -*-
"""
logging hooks module.
"""

import pyrin.logging.services as logging_services

from pyrin.packaging.decorators import packaging_hook
from pyrin.packaging.hooks import PackagingHookBase


@packaging_hook()
class PackagingHook(PackagingHookBase):
    """
    packaging hook class.
    """

    def __init__(self):
        """
        initializes an instance of PackagingHook.
        """

        super().__init__()

    def after_packages_loaded(self):
        """
        this method will be called after all application packages has been loaded.
        """

        # we must wrap all available loggers into an adapter
        # to inject request info into every log record.
        # it does not affect sqlalchemy logs.
        logging_services.wrap_all_loggers()



# -*- coding: utf-8 -*-
"""
security services module.
"""

from pyrin.application.services import get_component
from pyrin.security import SecurityPackage


def get_password_hash(password, **options):
    """
    gets the given password's hash.

    :param str password: password to get it's hash.

    :keyword bool is_encrypted: specifies that given password is encrypted.

    :raises InvalidPasswordLengthError: invalid password length error.

    :rtype: str
    """

    return get_component(SecurityPackage.COMPONENT_NAME).get_password_hash(password, **options)


def encrypt(text, **options):
    """
    encrypts the given text and returns the encrypted value.

    :param str text: text to be encrypted.

    :raises InvalidEncryptionTextLengthError: invalid encryption text length error.

    :rtype: str
    """

    return get_component(SecurityPackage.COMPONENT_NAME).encrypt(text, **options)


def has_permission(user, permissions, **options):
    """
    gets a value indicating that given user has the specified permissions.

    :param user: user identity to check its permissions.
    :param list[PermissionBase] permissions: permissions to check for user.

    :rtype: bool
    """

    return get_component(SecurityPackage.COMPONENT_NAME).has_permission(user, permissions,
                                                                        **options)



# -*- coding: utf-8 -*-
"""
utils sqlalchemy module.
"""

from sqlalchemy.util import lightweight_named_tuple

import pyrin.utils.datetime as datetime_utils

from pyrin.core.context import DTO
from pyrin.core.globals import LIST_TYPES
from pyrin.utils.exceptions import InvalidRowResultFieldsAndValuesError, \
    FieldsAndValuesCountMismatchError


LIKE_CHAR_COUNT_LIMIT = 50


def entity_to_dict(entity, exposed_only=True):
    """
    converts the given entity into a dict and returns it.
    the result dict only contains the columns of the entity
    which their `hidden` attribute is set to False.

    :param CoreEntity entity: entity to be converted.

    :param bool exposed_only: if set to False, it returns all
                              columns of the entity as dict.
                              if not provided, defaults to True.
    
    :rtype: dict
    """

    if entity is None:
        return DTO()

    return entity.to_dict(exposed_only)


def dict_to_entity(entity_class, **kwargs):
    """
    converts the given keyword arguments into
    an specified entity and returns it.

    :param type entity_class: the result entity class type.

    :rtype: CoreEntity
    """

    result = entity_class()
    result.from_dict(**kwargs)
    return result


def entity_to_dict_list(entities, exposed_only=True):
    """
    converts the given list of entities into a
    list of dicts and returns the result.

    :param list[CoreEntity] entities: list of entities.

    :param bool exposed_only: if set to False, it returns all
                              columns of the entity as dict.
                              if not provided, defaults to True.

    :returns list[dict]
    :rtype list
    """

    results = []
    if entities is None or len(entities) <= 0:
        return results

    for single_entity in entities:
        results.append(entity_to_dict(single_entity, exposed_only))

    return results


def keyed_tuple_to_dict(value):
    """
    converts the given `AbstractKeyedTuple` object into a dict.

    :param AbstractKeyedTuple value: value to be converted.

    :rtype: dict
    """

    if value is None or len(value) <= 0:
        return DTO()

    return DTO(zip(value.keys(), value))


def keyed_tuple_to_dict_list(values):
    """
    converts the given list of `AbstractKeyedTuple` objects into a list of dicts.

    :param list[AbstractKeyedTuple] values: values to be converted.

    :rtype: dict
    """

    results = []
    if values is None or len(values) <= 0:
        return results

    for single_value in values:
        results.append(keyed_tuple_to_dict(single_value))

    return results


def like_both(value, start='%', end='%'):
    """
    gets a copy of string with `%` or couple of `_` values
    attached to both ends of it to use in like operator.
    this method is intended to be used as callable for
    sqlalchemy operator methods.

    :param str value: value to be processed.

    :param str start: start place holder to be prefixed.
                      it could be `%` or couple of `_` values
                      for exact matching.
                      defaults to `%` if not provided.

    :param str end: end place holder to be appended.
                    it could be `%` or couple of `_` values
                    for exact matching.
                    defaults to `%` if not provided.

    :rtype: str
    """

    if value is None:
        return None

    value = like_prefix(value, start)
    value = like_suffix(value, end)

    return value


def like_prefix(value, start='%'):
    """
    gets a copy of string with `%` or couple of `_` values
    attached to beginning of it to use in like operator.
    this method is intended to be used as callable for
    sqlalchemy operator methods.

    :param str value: value to be processed.

    :param str start: start place holder to be prefixed.
                      it could be `%` or couple of `_` values
                      for exact matching.
                      defaults to `%` if not provided.

    :rtype: str
    """

    if value is None:
        return None

    return '{start}{value}'.format(start=start, value=value)


def like_suffix(value, end='%'):
    """
    gets a copy of string with `%` or couple of `_` values
    attached to end of it to use in like operator.
    this method is intended to be used as callable for
    sqlalchemy operator methods.

    :param str value: value to be processed.

    :param str end: end place holder to be appended.
                    it could be `%` or couple of `_` values
                    for exact matching.
                    defaults to `%` if not provided.

    :rtype: str
    """

    if value is None:
        return None

    return '{value}{end}'.format(value=value, end=end)


def _process_place_holder(value, count):
    """
    processes the value and generates a place holder with count
    of `_` chars. this value could be used in like operator.

    :param str value: value to be processed.
    :param int count: count of `_` chars to be attached.

    :note count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`, if
                 the provided value goes upper than this limit, a
                 `%` will be attached instead of it. this limit is
                 for security reason.

    :rtype: str
    """

    if value is None:
        return ''

    if count is None or count <= 0:
        return ''

    place_holder = None
    if count > LIKE_CHAR_COUNT_LIMIT:
        place_holder = '%'

    if place_holder is None:
        place_holder = '_' * count

    return place_holder


def like_exact_both(value, count):
    """
    gets a copy of string with `_` attached to both ends of
    it by count of underscores to use in like operator.

    :param str value: value to be processed.
    :param int count: count of `_` chars to be attached.

    :note count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`, if
                 the provided value goes upper than this limit, a
                 `%` will be attached instead of it. this limit is
                 for security reason.

    :rtype: str
    """

    place_holder = _process_place_holder(value, count)
    return like_both(value, place_holder, place_holder)


def like_exact_prefix(value, count):
    """
    gets a copy of string with `_` attached to beginning of
    it by count of underscores to use in like operator.

    :param str value: value to be processed.
    :param int count: count of `_` chars to be attached.

    :note count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`, if
                 the provided value goes upper than this limit, a
                 `%` will be attached instead of it. this limit is
                 for security reason.

    :rtype: str
    """

    place_holder = _process_place_holder(value, count)
    return like_prefix(value, place_holder)


def like_exact_suffix(value, count):
    """
    gets a copy of string with `_` attached to end of
    it by count of underscores to use in like operator.

    :param str value: value to be processed.
    :param int count: count of `_` chars to be attached.

    :note count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`, if
                 the provided value goes upper than this limit, a
                 `%` will be attached instead of it. this limit is
                 for security reason.

    :rtype: str
    """

    place_holder = _process_place_holder(value, count)
    return like_suffix(value, place_holder)


def add_range_clause(clauses, column, value_lower, value_upper,
                     include_equal_to_lower=True,
                     include_equal_to_upper=True):
    """
    adds range comparison into given clauses using specified inputs.

    :param list clauses: clause list to add range clause to it.
    :param CoreColumn column: entity column to add range clause for it.
    :param object value_lower: lower bound of range clause.
    :param object value_upper: upper bound of range clause.

    :param include_equal_to_lower: specifies that lower value
                                   should be considered in range.
                                   defaults to True if not provided.

    :param include_equal_to_upper: specifies that upper value
                                   should be considered in range.
                                   defaults to True if not provided.
    """

    if value_lower is None and value_upper is None:
        return

    # we swap the upper and lower values in case of user mistake.
    if value_lower is not None and value_upper is not None and value_lower > value_upper:
        value_lower, value_upper = value_upper, value_lower

    if value_lower is not None and value_lower == value_upper:
        clauses.append(column == value_lower)
    else:
        if value_lower is not None:
            if include_equal_to_lower is True:
                clauses.append(column >= value_lower)
            else:
                clauses.append(column > value_lower)
        if value_upper is not None:
            if include_equal_to_upper is True:
                clauses.append(column <= value_upper)
            else:
                clauses.append(column < value_upper)


def add_datetime_range_clause(clauses, column,
                              value_lower, value_upper,
                              include_equal_to_lower=True,
                              include_equal_to_upper=True,
                              **options):
    """
    adds datetime range comparison into given clauses using specified inputs.

    :param list clauses: clause list to add datetime range clause to it.
    :param CoreColumn column: entity column to add datetime range clause for it.
    :param datetime value_lower: lower bound of datetime range clause.
    :param datetime value_upper: upper bound of datetime range clause.

    :param include_equal_to_lower: specifies that lower datetime
                                   should be considered in range.
                                   defaults to True if not provided.

    :param include_equal_to_upper: specifies that upper datetime
                                   should be considered in range.
                                   defaults to True if not provided.

    :keyword bool consider_begin_of_day: specifies that consider begin
                                         of day for lower datetime.
                                         defaults to True if not provided.

    :keyword bool consider_end_of_day: specifies that consider end
                                       of day for upper datetime.
                                       defaults to True if not provided.
    """

    value_lower, value_upper = datetime_utils.normalize_datetime_range(value_lower,
                                                                       value_upper,
                                                                       **options)

    add_range_clause(clauses, column,
                     value_lower, value_upper,
                     include_equal_to_lower,
                     include_equal_to_upper)


def add_comparison_clause(clauses, column, value):
    """
    adds list or single comparison into clauses based on given value.
    if the value type is any of list, tuple or set, it generates an
    `in()` comparison, otherwise it generates a simple `==` comparison.

    :param list clauses: clause list to add comparison clause to it.
    :param CoreColumn column: entity column to add comparison clause for it.
    :param Union[object, list[object]] value: value to add comparison for it.
    """

    if value is not None:
        if isinstance(value, LIST_TYPES):
            clauses.append(column.in_(value))
        else:
            clauses.append(column == value)


def create_row_result(fields, values):
    """
    creates a row result object with given fields and values.
    this object type is returned by sqlalchemy `Query` when there
    is column names or multiple entities in query.

    :param list[str] fields: field names of the result object.

    :param list[object] values: values to be mapped to fields.
                                they must be in the same order as fields.

    :raises InvalidRowResultFieldsAndValuesError: invalid row result fields
                                                  and values error.

    :raises FieldsAndValuesCountMismatchError: fields and values count mismatch error.

    :rtype: AbstractKeyedTuple
    """

    if fields is None or values is None:
        raise InvalidRowResultFieldsAndValuesError('Input parameters "fields" and '
                                                   '"values" must both be provided, '
                                                   'they could not be None.')

    if len(fields) != len(values):
        raise FieldsAndValuesCountMismatchError('The length of "fields" which is '
                                                '[{fields}] and "values" which is '
                                                '[{values}] does not match.'
                                                .format(fields=len(fields),
                                                        values=len(values)))

    keyed_tuple = lightweight_named_tuple('result', fields)
    result = keyed_tuple(values)

    return result



# -*- coding: utf-8 -*-
"""
router package.
"""

from pyrin.packaging.context import Package


class RouterPackage(Package):
    """
    router package class.
    """

    NAME = __name__
    DEPENDS = []
    COMPONENT_NAME = 'api.router.component'



# -*- coding: utf-8 -*-
"""
application hooks module.
"""

from pyrin.core.context import Hook


class ApplicationHookBase(Hook):
    """
    application hook base class.
    all packages that need to be hooked into application business must
    implement this class and register it in application hooks.
    """

    def __init__(self):
        """
        initializes an instance of ApplicationHookBase.
        """

        super().__init__()

    def after_application_loaded(self):
        """
        this method will be called after application has been loaded.
        """
        pass

    def before_application_start(self):
        """
        this method will be called before application gets started.
        note that this method will not get called when application
        starts in scripting mode.
        """
        pass

    def application_status_changed(self, old_status, new_status):
        """
        this method will be called whenever application status changes.

        :param str old_status: old application status.
        :param str new_status: new application status.

        :note status:
            INITIALIZING = 'Initializing'
            LOADING = 'Loading'
            RUNNING = 'Running'
            TERMINATED = 'Terminated'
        """
        pass
